from __future__ import annotations

import json
import os
import pathlib
from dataclasses import dataclass
from time import time
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Set, Tuple
from warnings import warn

import img2pdf  # type: ignore
from typing_extensions import TypeAlias

from neoscore.core.brush import Brush, BrushDef
from neoscore.core.color import Color, ColorDef
from neoscore.core.exceptions import InvalidImageFormatError
from neoscore.core.key_event import KeyEvent
from neoscore.core.mouse_event import MouseEvent
from neoscore.core.paper import A4, Paper
from neoscore.core.pen import Pen
from neoscore.core.point import Point, PointDef
from neoscore.core.propagating_thread import PropagatingThread
from neoscore.core.rect import RectDef
from neoscore.core.units import Unit
from neoscore.interface.app_interface import AppInterface

if TYPE_CHECKING:
    from neoscore.core.document import Document
    from neoscore.core.font import Font


"""The global application state module."""

default_font: Font
"""The default font to be used in ``Text`` objects."""

document: Document
"""The root document object."""

registered_music_fonts: Dict[str, dict] = {}
"""A map from registered music font names to SMuFL metadata"""

registered_font_family_names: Set[str] = set()
"""A set of family names of all registered fonts, including music fonts"""

background_brush = Brush("#ffffff")
"""The brush used to draw the scene background.

Defaults to solid white. Set this using :obj:`.set_background_brush`.
"""

app_interface: AppInterface
"""The underlying application interface.

You generally shouldn't directly interact with this.
"""

_display_page_geometry_in_refresh_func: bool = False
"""Whether to render the page geometry in refresh renders.

This global serves as a bit of a hack for passing the "display page geometry?" flag into
refresh functions.
"""

_must_clear_scene_before_next_render: bool = False
"""Whether the scene must be cleared before rendering.

When refresh functions indicate no re-render is required, that indication takes
precedence over this flag.
"""

_supported_image_extensions = {
    ".bmp",
    ".jpg",
    ".jpeg",
    ".png",
    ".pbm",
    ".pgm",
    ".ppm",
    ".xbm",
    ".xpm",
}

# Directories
_FONTS_DIR = pathlib.Path(__file__).parent / ".." / "resources" / "fonts"

# Text Font
_LORA_DIR = _FONTS_DIR / "lora"
_DEFAULT_LORA_FONT_FAMILY_NAME = "Lora"
_DEFAULT_LORA_FONT_SIZE = Unit(12)
_LORA_REGULAR_PATH = _LORA_DIR / "Lora-VariableFont_wght.ttf"
_LORA_ITALIC_PATH = _LORA_DIR / "Lora-Italic-VariableFont_wght.ttf"

# Music Text Font
_BRAVURA_DIR = _FONTS_DIR / "bravura"
_BRAVURA_PATH = _BRAVURA_DIR / "Bravura.otf"
_BRAVURA_METADATA_PATH = _BRAVURA_DIR / "bravura_metadata.json"


def setup(paper: Paper = A4):
    """Initialize the application and set up the global state.

    This initializes the global ``Document`` and a back-end
    AppInterface instance.

    This should be called once at the beginning of every script using ``neoscore``;
    calling this multiple times in one script will cause unexpected behavior.

    Args:
        paper: The paper to use in the document.
    """
    global app_interface
    global default_font
    global document
    global background_brush
    # Some things are imported here to work around cyclic import problems
    from neoscore.core.document import Document
    from neoscore.core.font import Font

    document = Document(paper)

    app_interface = AppInterface(
        document, _repl_refresh_func, background_brush.interface, True
    )
    _register_default_fonts()
    default_font = Font(
        _DEFAULT_LORA_FONT_FAMILY_NAME, _DEFAULT_LORA_FONT_SIZE, 1, False
    )


def set_default_color(color: ColorDef):
    """Set the default color used in unspecified pens and brushes.

    This only affects objects created after this is called."""
    # Objects using unspecified pens and brushes make copies of these
    # global default objects.
    c = Color.from_def(color)
    Pen._default_color = c
    Brush._default_color = c


def set_background_brush(brush: BrushDef):
    """Set the brush used to paint the scene background."""
    global background_brush
    global app_interface
    background_brush = Brush.from_def(brush)
    app_interface.background_brush = background_brush.interface


def register_font(font_file_path: str | pathlib.Path) -> List[str]:
    """Register a font file with the application.

    If successful, this makes the font available for use in :obj:`.Font` objects,
    to be referenced by the family name embedded in the font file.

    Args:
        font_file_path: A path to a font file. Only TrueType and OpenType
            fonts are supported.

    Returns:
        A list of family names registered. Typically, this will have length 1.

    Raises:
        FontRegistrationError: If the font could not be loaded.
            Typically, this is because the given path does not lead to
            a valid font file.
    """
    global registered_font_family_names
    global app_interface
    family_names = app_interface.register_font(font_file_path)
    for name in family_names:
        registered_font_family_names.add(name)
    return family_names


def register_music_font(
    font_file_path: str | pathlib.Path, metadata_path: str | pathlib.Path
):
    """Register a music font with the application.

    Args:
        font_file_path: A path to a font file.
        metadata_path: A path to a SMuFL metadata JSON file
            for this font. The standard SMuFL format for this file name
            will be ``{lowercase_font_name}_metadata.json``.

    Returns:
        A list of family names registered. Typically, this will have length 1.

    Raises:
        FontRegistrationError: If the font could not be loaded.
            Typically, this is because the given path does not lead to
            a valid font file.
    """
    global registered_music_fonts
    family_names = register_font(font_file_path)
    try:
        with open(metadata_path, "r") as metadata_file:
            metadata = json.load(metadata_file)
    except FileNotFoundError:
        raise FileNotFoundError(
            "Music font metadata file {} could not be found".format(metadata_path)
        )
    except json.JSONDecodeError as e:
        e.msg = "Invalid JSON metadata in music font " "metadata file {}".format(
            metadata_path
        )
        raise e
    name = family_names[0]
    if len(family_names) > 1:
        print(
            f"Warning: music font at {font_file_path} contained more than 1 font "
            + f"family. SMuFL metadata will only be stored for {name}."
        )
    registered_music_fonts[name] = metadata


@dataclass
class RefreshFuncResult:
    """Results passed back to the neoscore runtime from refresh functions."""

    scene_render_needed: bool = True
    """If True, neoscore will clear the scene and re-render it.

    Refresh functions can set this to false to tell neoscore that no changes to the
    scene were made during the refresh, and so re-rendering is not necessary. This is a
    helpful optimization in situations where the scene does not change in every frame.
    """


RefreshFunc: TypeAlias = Callable[[float], Optional[RefreshFuncResult]]
"""A user-providable function for updating the scene every frame(ish).

The function should accept one argument - the current time in seconds.

Refresh functions can modify the scene, create new objects, and :obj:`remove
<.PositionedObject.remove>` them, though not all objects respond well to mutability.

The function may optionally return a :obj:`.RefreshFuncResult` to pass information back
to the neoscore runtime. If this is omitted, a default :obj:`.RefreshFuncResult` is
automatically returned.
"""


def show(
    refresh_func: Optional[RefreshFunc] = None,
    display_page_geometry=True,
    auto_viewport_interaction_enabled=True,
    min_window_size: Optional[Tuple[int, int]] = None,
    max_window_size: Optional[Tuple[int, int]] = None,
    fullscreen: bool = False,
):
    """Display the score in an interactive GUI window.

    Args:
        refresh_func: A scene update function to run on a timer approximating the
            frame rate. This can also be set with :obj:`.set_refresh_func`, which
            allows customizing the target frame rate.
        display_page_geometry: Whether to include a preview of page geometry,
            including a page outline and a dotted outline of the page's live
            area inside its margins.
        auto_viewport_interaction_enabled: Whether mouse and scrollbar viewport
            interaction is enabled. If false, scrollbars do not appear, mousewheel
            zooming is disabled, and click-and-drag view movement is disabled.
        min_window_size: An optional ``(width, height)`` minimum window size tuple.
        max_window_size: An optional ``(width, height)`` maximum window size tuple.
        fullscreen: Whether to show the window in fullscreen mode.
            This doesn't mix well with ``max_window_size``.
    """
    global app_interface
    global background_brush
    global _display_page_geometry_in_refresh_func
    _display_page_geometry_in_refresh_func = display_page_geometry

    _render_document(display_page_geometry, background_brush)
    if refresh_func:
        set_refresh_func(refresh_func)
    app_interface.auto_viewport_interaction_enabled = auto_viewport_interaction_enabled
    app_interface.show(min_window_size, max_window_size, fullscreen)


def _render_document(display_page_geometry: bool, background_brush: Brush):
    """Render the document, clearing the scene before if needed.

    This should be used instead of using ``document.render`` directly.
    """
    global document
    global app_interface
    global _must_clear_scene_before_next_render

    if _must_clear_scene_before_next_render:
        app_interface.clear_scene()
        for page in document.pages:
            for obj in page.descendants:
                interfaces = getattr(obj, "interfaces", None)
                if interfaces:
                    interfaces.clear()
                if hasattr(obj, "_interface_for_children"):
                    obj._interface_for_children = None
    document.render(display_page_geometry, background_brush)
    _must_clear_scene_before_next_render = True


def set_viewport_center_pos(document_pos: PointDef):
    """Center the interactive viewport at a given document-space position.

    If the given position is such that the requested view goes beyond the document's
    bounding rect, the actual position used is adjusted to prevent that. This is caused
    by an internal limitation and may change in the future. To work around this issue,
    you can expand the document bounding rect by creating an arbitrarily large invisible
    path like so::

        Path.rect((Mm(-1000000), Mm(-1000000)), None, Mm(2000000), Mm(2000000),
              Brush.no_brush(), Pen.no_pen())

    This is mostly recommended when not using automatic viewport interaction.
    """
    global app_interface
    app_interface.viewport_center_pos = Point.from_def(document_pos)


def get_viewport_center_pos() -> Point:
    """Return the interactive viewport's center position in document-space.

    This value may differ slightly from values set in ``set_viewport_center_pos``.
    """
    global app_interface
    return app_interface.viewport_center_pos


def set_viewport_scale(scale: float):
    """Set the interactive viewport's scale (zoom).

    Values should be greater than 0, with 1 as the base zoom and
    larger numbers zooming in.
    """
    global app_interface
    app_interface.viewport_scale = scale


def get_viewport_scale() -> float:
    """Return the interactive viewport's scale (zoom).

    This value is always greater than 0, with 1 as the base zoom and
    larger numbers zooming in.

    This value may differ slightly from values set in ``set_viewport_scale``.
    """
    global app_interface
    return app_interface.viewport_scale


def set_viewport_rotation(rotation: float):
    """Set the interactive viewport's rotation angle in degrees.

    The viewport is rotated about its center.
    """
    global app_interface
    app_interface.viewport_rotation = rotation


def get_viewport_rotation() -> float:
    """Return the interactive viewport's rotation angle in degrees."""
    global app_interface
    return app_interface.viewport_rotation


def render_pdf(pdf_path: str | pathlib.Path, dpi: int = 300):
    """Render the score as a pdf.

    Args:
        pdf_path: The output pdf path
        dpi: Resolution to render at
    """
    global app_interface
    global background_brush
    _render_document(False, background_brush)
    # Render all pages to temp files
    page_imgs = []
    render_threads = []
    for page in document.pages:
        img_buffer = bytearray()
        page_imgs.append(img_buffer)
        render_threads.append(
            render_image(
                page.document_space_bounding_rect,
                img_buffer,
                dpi,
                preserve_alpha=False,
                wait=False,
            )
        )
    for thread in render_threads:
        thread.join()
    # Assemble into PDF and write it to file path
    with open(pdf_path, "wb") as f:
        f.write(img2pdf.convert([bytes(buf) for buf in page_imgs]))


def render_image(
    rect: Optional[RectDef],
    dest: str | pathlib.Path | bytearray,
    dpi: int = 300,
    quality: int = -1,
    autocrop: bool = False,
    preserve_alpha: bool = True,
    wait: bool = True,
) -> PropagatingThread:
    """Render a section of the document to an image.

    The following file extensions are supported:

        * ``.bmp``
        * ``.jpg``
        * ``.png``
        * ``.pbm``
        * ``.pgm``
        * ``.ppm``
        * ``.xbm``
        * ``.xpm``

    If `wait == False`, this renders on the main thread but autocrops and saves the
    image on a spawned thread which is returned to allow efficient rendering of many
    images in parallel.

    Args:
        rect: The part of the document to render, in document coordinates.
            If ``None``, the entire scene will be rendered.
        dest: An output file path or a bytearray to save to. If a bytearray
            is given, the output format will be PNG.
        dpi: The pixels per inch of the rendered image. quality: The quality of the
        output image for compressed
            image formats. Must be either ``-1`` (default compression) or between ``0``
            (most compressed) and ``100`` (least compressed).
        autocrop: Whether to crop the output image to tightly
            fit the contents of the frame.
        preserve_alpha: Whether to preserve the alpha channel. If false,
            ``neoscore.background_brush`` will be used to flatten any transparency.
        wait: Whether to block until the image is fully exported.

    Raises:
        InvalidImageFormatError: If the given ``image_path`` does not have a
            supported image format file extension.
        ImageExportError: If low level Qt image export fails for
            unknown reasons.
    """

    global app_interface
    global background_brush

    if not ((0 <= quality <= 100) or quality == -1):
        warn("render_image quality {} invalid; using default.".format(quality))
        quality = -1

    if (
        not isinstance(dest, bytearray)
        and not os.path.splitext(dest)[1] in _supported_image_extensions
    ):
        raise InvalidImageFormatError(
            "image_path {} is not in a supported format.".format(dest)
        )

    bg_color = background_brush.color
    _render_document(False, background_brush)

    thread = app_interface.render_image(
        rect,
        dest,
        dpi,
        quality,
        bg_color,
        autocrop,
        preserve_alpha,
    )
    if wait:
        thread.join()
    return thread


def _repl_refresh_func(_: float) -> float:
    """Default refresh func to be used in REPL mode.

    Refreshes at a rate of 5 FPS.
    """
    global background_brush
    global _display_page_geometry_in_refresh_func
    _render_document(_display_page_geometry_in_refresh_func, background_brush)
    return 0.2


def set_refresh_func(refresh_func: RefreshFunc, target_fps: int = 60):
    """Update the global scene refresh function.

    Args:
        refresh_func: The new refresh function
        target_fps: The requested frame rate to run the function at.
    """
    global app_interface
    global background_brush
    global _display_page_geometry_in_refresh_func

    frame_wait = 1 / target_fps

    # Wrap the user-provided refresh function with code that clears
    # the scene and re-renders it, then returns the requested delay
    # before the next frame, calculated to automatically compensate
    # for refresh time.
    def wrapped_refresh_func(frame_time: float) -> float:
        result = refresh_func(frame_time)
        if result is None:
            # Construct default result if none was provided
            result = RefreshFuncResult()
        if result.scene_render_needed:
            _render_document(_display_page_geometry_in_refresh_func, background_brush)
        elapsed_time = time() - frame_time
        return max(frame_wait - elapsed_time, 0)

    app_interface.set_refresh_func(wrapped_refresh_func)


def set_mouse_event_handler(handler: Callable[[MouseEvent], None]):
    """Set the global mouse event handler function.

    Mouse events are fired on mouse moves and button presses, releases, and double
    clicks.

    Args:
        handler: A callback function which accepts a mouse event.
            See :obj:`.MouseEvent` for what event data is provided.

    """
    global app_interface
    app_interface.set_mouse_event_handler(handler)


def set_key_event_handler(handler: Callable[[KeyEvent], None]):
    """Set the global key event handler function.

    Key events are fired on key presses, releases, and auto-repeats from being held.

    Args:
        handler: A callback function which accepts a key event.
            See :obj:`.KeyEvent` for what event data is provided.

    """
    global app_interface
    app_interface.set_key_event_handler(handler)


def _register_default_fonts():
    register_music_font(
        _BRAVURA_PATH,
        _BRAVURA_METADATA_PATH,
    )
    register_font(_LORA_REGULAR_PATH)
    register_font(_LORA_ITALIC_PATH)


def shutdown():
    """Tear down the global state, including the document tree.

    After running this, :obj:`.neoscore.setup` can be called again to start a new
    document.
    """
    app_interface.destroy()



from typing import Optional

from neoscore.core.break_opportunity import BreakOpportunity
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject


class BreakHint(PositionedObject, BreakOpportunity):

    """A standalone invisible break opportunity.

    When placed in a :obj:`.Flowable`, this signals to the flowable that a break can be
    performed here.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
    ):
        """
        Args:
            pos: The position of the object relative to its parent
            parent: The parent object.
        """
        PositionedObject.__init__(self, pos, parent)
        BreakOpportunity.__init__(self)



from __future__ import annotations

from typing import Tuple, Union

from typing_extensions import TypeAlias

from neoscore.core.exceptions import ColorBoundsError


class Color:
    """An 8-bit int RGBA color."""

    def __init__(self, *args):
        """
        Valid signatures:
            * Color(css_hex_string)
            * Color(red, green, blue)
            * Color(red, green, blue, alpha)
        """
        if len(args) == 1:
            self._set_with_hex(*args)
        elif len(args) == 3:
            self._set_with_rgb(*args)
        elif len(args) == 4:
            self._set_with_rgba(*args)
        else:
            raise TypeError("Invalid Color init args")

    @classmethod
    def from_def(cls, color_def: ColorDef) -> Color:
        if isinstance(color_def, Color):
            return color_def
        elif isinstance(color_def, tuple):
            return Color(*color_def)
        else:
            # otherwise color_def must be a str
            return Color(color_def)

    def __repr__(self):
        return "{}({}, {}, {}, {})".format(
            type(self).__name__, self.red, self.green, self.blue, self.alpha
        )

    def __eq__(self, other):
        """Two Colors are considered equal if all of their properties are."""
        return (
            type(other) == type(self)
            and self.red == other.red
            and self.green == other.green
            and self.blue == other.blue
            and self.alpha == other.alpha
        )

    def __hash__(self):
        """Colors with equal properties will have the same hash."""
        return 23467817 ^ self.red ^ self.green ^ self.blue ^ self.alpha

    @property
    def red(self):
        """int: The 0-255 value of the red color channel"""
        return self._red

    @property
    def green(self):
        """int: The 0-255 value of the green color channel"""
        return self._green

    @property
    def blue(self):
        """int: The 0-255 value of the blue color channel"""
        return self._blue

    @property
    def alpha(self):
        """int: The 0-255 value of the alpha color channel"""
        return self._alpha

    def _set_with_hex(self, hex_value):
        """Set properties from an #rrggbb hex string

        Args:
            hex_value (str): A CSS hex color string, with or without a leading '#'

        Returns: None
        """
        if hex_value.startswith("#"):
            hex_value = hex_value[1:]

        # Short or long form Hex
        if 3 <= len(hex_value) <= 4:
            self._red = int(hex_value[0] + (hex_value[0]), 16)
            self._green = int(hex_value[1] + (hex_value[1]), 16)
            self._blue = int(hex_value[2] + (hex_value[2]), 16)
            if len(hex_value) == 4:
                self._alpha = int(hex_value[3] + (hex_value[3]), 16)
            else:
                self._alpha = 255
        else:
            self._red = int(hex_value[0:2], 16)
            self._green = int(hex_value[2:4], 16)
            self._blue = int(hex_value[4:6], 16)
            if len(hex_value) == 8:
                self._alpha = int(hex_value[6:8], 16)
            else:
                self._alpha = 255

        self._validate_channel_values()

    def _set_with_rgb(self, red, green, blue):
        """Set properties from three 0-255 red, green, and blue values

        Args:
            red (int): A 0-255 red channel value
            green (int): A 0-255 green channel value
            blue (int): A 0-255 blue channel value

        Returns: None
        """
        self._red = red
        self._green = green
        self._blue = blue
        self._alpha = 255
        self._validate_channel_values()

    def _set_with_rgba(self, red, green, blue, alpha):
        """Set properties from four 0-255 red, green, blue, and alpha values

        Args:
            red (int): A 0-255 red channel value
            green (int): A 0-255 green channel value
            blue (int): A 0-255 blue channel value
            alpha (int): A 0-255 alpha channel value

        Returns: None
        """
        self._red = red
        self._green = green
        self._blue = blue
        self._alpha = alpha
        self._validate_channel_values()

    def _validate_channel_values(self):
        for value in [self._red, self._green, self._blue, self._alpha]:
            if not (0 <= value <= 255):
                raise ColorBoundsError(value)


ColorDef: TypeAlias = Union[Color, str, Tuple[int, int, int], Tuple[int, int, int, int]]
"""A ``Color`` or a shorthand hex string or init argument tuple for one.

Hex strings may have 3 channels (#RRGGBB) or 4 (#RRGGBBAA).
"""



from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from fractions import Fraction
from typing import Optional, Tuple, Union

from typing_extensions import TypeAlias

from neoscore.core.math_helpers import is_power_of_2
from neoscore.western.duration_display import DurationDisplay


@dataclass(frozen=True)
class Duration:
    """A metered non-tuplet duration.

    The duration fraction indicates duration as a fraction of a whole note.
    The actual written denomination of duration is deduced
    from the reduced fraction. For instance:

    * ``Duration(1, 4)`` indicates a quarter note value
    * ``Duration(1, 1)`` indicates a whole note value
    * ``Duration(3, 8)`` indicates a dotted quarter note value

    """

    numerator: InitVar[int]

    denominator: InitVar[int]

    fraction: Fraction = field(init=False)
    """The reduced fraction representation of the duration"""

    display: Optional[DurationDisplay] = field(init=False, compare=False)
    """The appearance spec of the duration when written in notes or rests.

    This is ``None`` if the duration cannot be represented without ties.
    """

    def __post_init__(self, numerator: int, denominator: int):
        if numerator <= 0:
            raise ValueError("Numerator must be positive")
        if not is_power_of_2(denominator):
            raise ValueError("Denominator must be a power of 2")
        super().__setattr__("fraction", Fraction(numerator, denominator))
        super().__setattr__("display", Duration._derive_display(self.fraction))

    @staticmethod
    def _derive_display(fraction: Fraction) -> Optional[DurationDisplay]:
        if fraction >= 2:
            if fraction >= 4:
                # Lengths >= 4 whole notes cannot be represented without a tie
                return None
            # The below algorithm doesn't work with double-breves, so
            # hack it by running it on fraction / 2, giving a
            # whole-note value, then plugging its dot count into the
            # result with double-breve (base_division 0) hardcoded
            sub_result = Duration._derive_display(fraction / 2)
            if sub_result is None:
                return None
            assert sub_result.base_duration == 1
            return DurationDisplay(0, sub_result.dot_count)
        partial_numerator: float = fraction.numerator
        partial_denominator: int = fraction.denominator
        dot_count = 0
        while partial_numerator > 1:
            partial_numerator = (partial_numerator - 1) / 2
            partial_denominator = partial_denominator // 2
            dot_count += 1
        if partial_numerator != 1:
            # Failure to reduce means the duration requires a tie to write
            return None
        return DurationDisplay(partial_denominator, dot_count)

    @classmethod
    def from_def(cls, duration_def: DurationDef) -> Duration:
        if isinstance(duration_def, Duration):
            return duration_def
        return Duration(*duration_def)

    @classmethod
    def from_description(cls, base_division: int, dots: int) -> Duration:
        """Create a ``Duration`` from a base division and a number of dots.

        ``Duration``\ s created with this will always have a valid ``DurationDisplay``.

        Args:
            base_division: Must be 0 (double whole note) or a power of 2
            dots: Must be >= 0
        """
        if base_division == 0:
            # double breve
            val = Fraction(2, 1)
            inc_division = 1
        elif is_power_of_2(base_division):
            val = Fraction(1, base_division)
            inc_division = base_division * 2
        else:
            raise ValueError("base_division must be 0 or a power of 2")
        for _ in range(dots):
            val += Fraction(1, inc_division)
            inc_division *= 2
        return Duration(val.numerator, val.denominator)

    @property
    def requires_tie(self) -> bool:
        """If this Duration requires a tie to be written."""
        return self.display is None

    def __float__(self):
        """Reduce the fractional representation to a ``float`` and return it."""
        return float(self.fraction)

    def __add__(self, other: Duration):
        """Durations are added by adding their fractions."""
        if not isinstance(other, type(self)):
            raise TypeError
        fraction_sum = self.fraction + other.fraction
        return Duration(fraction_sum.numerator, fraction_sum.denominator)

    def __sub__(self, other: Duration):
        """Durations are subtracted by subtracting their fractions."""
        if not isinstance(other, type(self)):
            raise TypeError
        fraction_diff = self.fraction - other.fraction
        return Duration(fraction_diff.numerator, fraction_diff.denominator)

    def __gt__(self, other: Duration):
        """Durations are compared by their fractions."""
        if not isinstance(other, type(self)):
            return False
        return self.fraction > other.fraction

    def __ge__(self, other: Duration):
        """Durations are compared by their fractions."""
        return self > other or self == other

    def __lt__(self, other: Duration):
        """Durations are ordered by their fractions."""
        if not isinstance(other, type(self)):
            return False
        return self.fraction < other.fraction

    def __le__(self, other: Duration):
        """Durations are compared by their fractions."""
        return self < other or self == other


DurationDef: TypeAlias = Union[Duration, Tuple[int, int]]
"""A Duration or a shorthand tuple for one."""



from __future__ import annotations

from typing import TYPE_CHECKING

from backports.cached_property import cached_property

from neoscore.core.brush import Brush
from neoscore.core.color import Color
from neoscore.core.directions import DirectionX
from neoscore.core.paper import Paper
from neoscore.core.pen import Pen
from neoscore.core.pen_pattern import PenPattern
from neoscore.core.point import Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.rect import Rect
from neoscore.core.units import ZERO, Mm, Unit

if TYPE_CHECKING:
    from neoscore.core.document import Document


_PREVIEW_OUTLINE_COLOR = Color("#551155")
_PREVIEW_SHADOW_COLOR = Color(0, 0, 0, 80)


class Page(PositionedObject):

    """A document page.

    All :obj:`.PositionedObject`\ s will have a ``Page`` as their
    ancestor. All ``Page``\ s are direct children of the global document.

    ``Page`` objects are automatically created by ``Document`` and should
    not be manually created or manipulated.
    """

    _neoscore_page_type_marker = True

    def __init__(
        self,
        pos: PointDef,
        document: Document,
        index: int,
        page_side: DirectionX,
        paper: Paper,
    ):
        """
        Args:
            pos: The position of the top left corner of this page's live area
                in canvas space. Note that this refers to the corner of the *live*
                area bounded by margins, not the actual paper corner.
            document: The global document. This is used as the ``Page`` object's parent.
            index: The index of this page. This should be the same index this page
                can be found at in :obj:`.Document.pages`.
            page_side: The left/right side the page lies on when printed.
            paper: The paper geometry for this page.
        """
        super().__init__(pos, document)  # noqa
        self._document = document
        self._index = index
        self._page_side = page_side
        self.paper = paper
        self._geometry_preview_created = False

    @property
    def index(self):
        """The index of this page in its managing :obj:`.PageSupplier`."""
        return self._index

    @property
    def page_side(self):
        """The left/right side the page lies on when printed.

        This determines which side the gutter should be placed on.
        """
        return self._page_side

    @cached_property
    def bounding_rect(self) -> Rect:
        """The page bounding rect, positioned relative to the ``pos``."""
        if self.page_side == DirectionX.RIGHT:
            # Page is on right side, apply gutter on left side
            rect_x = -(self.paper.gutter + self.paper.margin_left)
        else:
            rect_x = -self.paper.margin_left
        return Rect(
            rect_x,
            -self.paper.margin_top,
            self.paper.width,
            self.paper.height,
        )

    @cached_property
    def document_space_bounding_rect(self) -> Rect:
        """The page bounding rect relative to the document."""
        local_rect = self.bounding_rect
        return Rect(
            local_rect.x + self.x,
            local_rect.y + self.y,
            local_rect.width,
            local_rect.height,
        )

    @cached_property
    def full_margin_left(self) -> Unit:
        """The left margin, including any gutter if the gutter is on the left."""
        if self.page_side == DirectionX.RIGHT:
            return self.paper.margin_left + self.paper.gutter
        else:
            return self.paper.margin_left

    @cached_property
    def full_margin_right(self) -> Unit:
        """The right margin, including any gutter if the gutter is on the right."""
        if self.page_side == DirectionX.RIGHT:
            return self.paper.margin_right
        else:
            return self.paper.margin_right + self.paper.gutter

    @property
    def left_margin_x(self) -> Unit:
        """The X position of the edge of the left margin.

        This is always ``ZERO``, given here as a convenience synonym.
        """
        return ZERO

    @property
    def right_margin_x(self) -> Unit:
        """The X position of the edge of the right margin.

        This is always ``page.paper.live_width``, given here as a convenience synonym.
        """
        return self.paper.live_width

    @property
    def top_margin_y(self) -> Unit:
        """The Y position of the edge of the top margin.

        This is always ``ZERO``, given here as a convenience synonym.
        """
        return ZERO

    @property
    def bottom_margin_y(self) -> Unit:
        """The Y position of the edge of the top margin.

        This is always ``page.paper.live_height``, given here as a convenience synonym.
        """
        return self.paper.live_height

    @property
    def center_x(self) -> Unit:
        """The X position of the center of the live page area.

        This is a convenience method for ``page.paper.live_width / 2``.
        """
        return self.paper.live_width / 2

    def create_geometry_preview(self, background_brush: Brush):
        """Create and render child objects which show the page geometry.

        This shouldn't be called directly; use the setting in :obj:`.neoscore.score`
        instead.

        This is useful for interactive views, but should typically not be called in PDF
        and image export contexts.
        """
        # An implementation note on page geometry previews:

        # Ideally it would make the most sense for page geometry preview objects to be
        # generated at the same time the page is created. However, this goal conflicts
        # with our desire for preview visibility to be set at the top-level
        # `neoscore.show()` call, typically at the end of a user script. This allows
        # users to switch between preview mode and pdf/image export without modifying
        # their setup line. Achieving this requires some ugly workarounds, like passing
        # preview information to the top-level `document.render()` call, and here
        # ensuring this is only ever called once with a private flag. There are more
        # elegant ways to approach this issue, but for now this seems ok.

        # Import here to avoid cyclic import
        from neoscore.core.path import Path

        # Ensure this is only executed once
        if self._geometry_preview_created:
            return
        self._geometry_preview_created = True

        # To ensure these preview objects appear below all real document objects, we
        # need to attach all preview objects to the *first* page. This is necessary
        # because the first page is drawn first. We furthermore need to ensure page
        # preview objects are placed at the beginning of the first page's child list to
        # ensure they are drawn before other first-page children. Some strange hacks
        # here are needed to guarantee this behavior.
        parent = self if self.index == 0 else None
        bounding_rect = self.bounding_rect
        page_drop_shadow_rect = Path.rect(
            self.pos + Point(bounding_rect.x + Mm(1), bounding_rect.y + Mm(1)),
            parent,
            bounding_rect.width,
            bounding_rect.height,
            Brush(_PREVIEW_SHADOW_COLOR),
            Pen.no_pen(),
        )
        page_preview_rect = Path.rect(
            self.pos + Point(bounding_rect.x, bounding_rect.y),
            parent,
            bounding_rect.width,
            bounding_rect.height,
            background_brush,
            pen=Pen(_PREVIEW_OUTLINE_COLOR),
        )
        live_area_preview_rect = Path.rect(
            self.pos,
            parent,
            self.paper.live_width,
            self.paper.live_height,
            Brush.no_brush(),
            pen=Pen(_PREVIEW_OUTLINE_COLOR, pattern=PenPattern.DOT),
        )
        # Terrible hack to ensure preview is always drawn below document contents
        preview_objs_parent = live_area_preview_rect.parent
        preview_objs_parent.children = (
            preview_objs_parent.children[-3:] + preview_objs_parent.children[:-3]
        )



import pathlib
from dataclasses import dataclass

from PyQt5.QtGui import QPixmap
from PyQt5.QtSvg import QGraphicsSvgItem
from PyQt5.QtWidgets import QGraphicsItem, QGraphicsPixmapItem

from neoscore.core.exceptions import ImageLoadingError
from neoscore.core.point import ORIGIN
from neoscore.interface.positioned_object_interface import PositionedObjectInterface
from neoscore.interface.qt.converters import point_to_qt_point_f

# Qt::SmoothTransformation
_QT_SMOOTH_TRANSFORMATION = 1


@dataclass(frozen=True)
class ImageInterface(PositionedObjectInterface):

    """Interface for images, including both pixmaps and SVGs.

    Supported image extensions/formats include: BMP, GIF, JPG, JPEG,
    PNG, PBM, PGM, PPM, XBM, XPM, and SVG.

    Scaling respects the image's aspect ratio, and is performed using
    bilinear filtering.
    """

    file_path: pathlib.Path
    """Path to an image file to be used"""

    opacity: float = 1
    """The image's opacity, where 1 is fully opaque and 0 is invisible."""

    @property
    def _resolved_path(self) -> str:
        return str(self.file_path.expanduser())

    def _create_svg_qt_object(self) -> QGraphicsSvgItem:
        qt_object = QGraphicsSvgItem(self._resolved_path)
        self._apply_common_properties(qt_object)
        return qt_object

    def _create_pixmap_qt_object(self) -> QGraphicsPixmapItem:
        pixmap = QPixmap()
        load_success = pixmap.load(self._resolved_path)
        if not load_success:
            raise ImageLoadingError(f"Failed to load image at {self.file_path}")
        qt_object = QGraphicsPixmapItem(pixmap)
        qt_object.setTransformationMode(_QT_SMOOTH_TRANSFORMATION)
        self._apply_common_properties(qt_object)
        return qt_object

    def _apply_common_properties(self, qt_object: QGraphicsItem):
        qt_object.setPos(point_to_qt_point_f(self.pos))
        if self.transform_origin != ORIGIN:
            qt_object.setTransformOriginPoint(
                point_to_qt_point_f(self.transform_origin)
            )
        if self.scale != 1:
            qt_object.setScale(self.scale)
        if self.rotation != 0:
            qt_object.setRotation(self.rotation)
        if self.transform_origin != ORIGIN:
            qt_object.setTransformOriginPoint(
                point_to_qt_point_f(self.transform_origin)
            )
        if self.opacity != 1:
            qt_object.setOpacity(self.opacity)

    def render(self):
        if self.file_path.suffix == ".svg":
            qt_object = self._create_svg_qt_object()
        else:
            qt_object = self._create_pixmap_qt_object()
        self._register_qt_object(qt_object)



"""Objects which control flowable layouts"""
from typing import cast

from neoscore.core.page import Page
from neoscore.core.point import Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import Unit


class LayoutController:
    """An abstract layout controller for working with Flowable layouts."""

    def __init__(self, flowable_x: Unit):
        """
        Args:
            flowable_x: The position of this controller within the owning
                flowable's local space.
        """
        self._flowable_x = flowable_x

    @property
    def flowable_x(self) -> Unit:
        """The x position in the flowable's local space."""
        return self._flowable_x


class NewLine(LayoutController, PositionedObject):
    """A line break controller.

    These are currently used only for automatically generated line breaks and should
    generally not be constructed directly. Line breaks can be controlled using
    :obj:`.BreakOpportunity` objects instead.
    """

    def __init__(
        self,
        pos: PointDef,
        page: Page,
        flowable_x: Unit,
        length: Unit,
        height: Unit,
    ):
        """
        Args:
            pos: The position of the top left corner of this line relative
                to the page.
            page: The page this line appears on. This is used as
                the object's parent.
            flowable_x: The x position in the flowable's local space where this
                line begins.
            length: The line length
            height: The line height
        """
        LayoutController.__init__(self, flowable_x)
        PositionedObject.__init__(self, pos, page)
        self._length = length
        self._height = height

    @property
    def page(self) -> Page:
        """The page this controller appears on.

        This is identical to :obj:`parent <.PositionedObject.parent>`.
        """
        return cast(Page, self.parent)

    @property
    def doc_end_pos(self) -> Point:
        """The position of the new line's bottom right corner.

        This position is relative to the page.
        """
        return Point(self.pos.x + self.length, self.pos.y + self.height)

    @property
    def length(self) -> Unit:
        """The length of the line."""
        return self._length

    @length.setter
    def length(self, value: Unit):
        self._length = value

    @property
    def height(self) -> Unit:
        """The height of the line."""
        return self._height

    @height.setter
    def height(self, value: Unit):
        self._height = value


class MarginController(LayoutController):
    """A controller defining flowable line margins.

    A flowable can have any number of different margin layers identified by a given
    ``layer_key``. A margin controller overrides the active margin values only for the
    specified layer. When a flowable generates its layout, it sums the margin values
    from all layers at the point of each new line to determine its margins.

    This layered margin system is useful for situations like staves, where different
    object states will change the required margins. For example, instrument names call
    for a margin size that will often be longer on the first system than later ones,
    while clefs and key signatures contribute their own margin amounts which can vary
    between systems. In this case, one might create separate margin controllers for each
    of these layers.

    These can be manually created and added to a flowable via
    :obj:`.Flowable.add_margin_controller` to control margins.
    """

    def __init__(self, flowable_x: Unit, margin_left: Unit, layer_key: str = ""):
        """
        Args:
            flowable_x: The position of this controller within the owning
                flowable's local space.
            margin_left: The margin value contributed by this controller in
                the specified layer
            layer_key: The layer this controller applies to. Layer keys starting with
                ``_neoscore`` are reserved for internal purposes.
        """
        super().__init__(flowable_x)
        self._margin_left = margin_left
        self._layer_key = layer_key

    @property
    def margin_left(self) -> Unit:
        """The left margin for generated lines while this controller is active"""
        return self._margin_left

    @property
    def layer_key(self) -> str:
        return self._layer_key



from __future__ import annotations

from typing import Optional

from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.western.accidental_type import AccidentalType


class Accidental(MusicText):

    """A visual accidental."""

    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        accidental_type: AccidentalType | str,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            pos: Position relative to ``parent``
            parent: If no font is given, this or one of its ancestors must
                implement :obj:`.HasMusicFont`.
            accidental_type: Which accidental to draw. For extended accidentals,
                an arbitrary string SMuFL glyph name may be provided.
            font: If provided, this overrides any font found in the ancestor chain.
        """
        self._accidental_type = accidental_type
        if isinstance(accidental_type, AccidentalType):
            canonical_name = accidental_type.value
        else:
            canonical_name = accidental_type
        MusicText.__init__(self, pos, parent, canonical_name, font)

    @property
    def accidental_type(self) -> AccidentalType | str:
        """The accidental variant.

        Can be set to either a standard accidental type or an
        arbitrary SMuFL glyph name.

        Setting this automatically updates the displayed glyph.
        """
        return self._accidental_type

    @accidental_type.setter
    def accidental_type(self, value: AccidentalType | str):
        self._accidental_type = value
        if isinstance(value, AccidentalType):
            self.text = value.value
        else:
            self.text = value



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type, cast

from neoscore.core.exceptions import NoClefError
from neoscore.core.layout_controllers import MarginController, NewLine
from neoscore.core.music_font import MusicFont
from neoscore.core.pen import Pen, PenDef
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject, render_cached_property
from neoscore.core.units import ZERO, Mm, Unit, make_unit_class
from neoscore.western.abstract_staff import AbstractStaff
from neoscore.western.staff_fringe_layout import StaffFringeLayout
from neoscore.western.staff_group import StaffGroup

if TYPE_CHECKING:
    from neoscore.western.clef import Clef
    from neoscore.western.key_signature import KeySignature


class Staff(AbstractStaff):
    """A staff with some high-level knowledge of its contents."""

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        length: Unit,
        group: Optional[StaffGroup] = None,
        line_spacing: Unit = Mm(1.75),
        line_count: int = 5,
        music_font_family: str = "Bravura",
        pen: Optional[PenDef] = None,
    ):
        """
        Args:
            pos: The position of the top-left corner of the staff
            parent: The parent for the staff. Make this a :obj:`.Flowable`
                to allow the staff to run across line and page breaks.
            length: The horizontal width of the staff
            group: The staff group this belongs to. Set this if being used in a system
                of multiple staves.
            line_spacing: The distance between two lines in the staff.
            line_count: The number of lines in the staff.
            music_font_family: The name of the font to use for MusicText objects
                in the staff. This defaults to the system-wide default music font
                family.
            pen: The pen used to draw the staff lines. Defaults to a line with
                thickness from the music font's engraving default.
        """
        unit = self._make_unit_class(line_spacing)
        music_font = MusicFont(music_font_family, unit)
        pen = pen or Pen(thickness=music_font.engraving_defaults["staffLineThickness"])
        super().__init__(
            pos, parent, length, group, unit(1), line_count, music_font, pen
        )

    def distance_to_next_of_type(self, staff_object: PositionedObject) -> Unit:
        """Find the x distance until the next occurrence of an object's type.

        If the object is the last of its type, this gives the remaining length of the
        staff after the object.

        This is useful for determining rendering behavior of staff objects which are
        active until another of their type occurs, such as :obj:`.KeySignature` and
        :obj:`.Clef`.
        """
        start_x = self.map_x_to(cast(PositionedObject, staff_object))
        all_others_of_class = (
            item
            for item in self.descendants_of_exact_class(type(staff_object))
            if item != staff_object
        )
        closest_x = Unit(float("inf"))
        for item in all_others_of_class:
            relative_x = self.map_x_to(item)
            if start_x < relative_x < closest_x:
                closest_x = relative_x
        if closest_x == Unit(float("inf")):
            return self.breakable_length - start_x
        return closest_x - start_x

    @render_cached_property
    def clefs(self) -> List[Tuple[Unit, Clef]]:
        """All the clefs in this staff, ordered by their relative x pos."""
        return self.find_ordered_descendants_with_attr("middle_c_staff_position")

    def active_clef_at(self, pos_x: Unit) -> Optional[Clef]:
        """Return the active clef at a given x position, if any."""
        return next(
            (clef for (clef_x, clef) in reversed(self.clefs) if clef_x <= pos_x),
            None,
        )

    @render_cached_property
    def key_signatures(self) -> List[Tuple[Unit, KeySignature]]:
        """All the key signatures in this staff, ordered by their relative x pos."""
        return self.find_ordered_descendants_with_attr(
            "_neoscore_key_signature_type_marker"
        )

    @render_cached_property
    def time_signatures(self) -> List[Tuple[Unit, KeySignature]]:
        """All the time signatures in this staff, ordered by their relative x pos."""
        return self.find_ordered_descendants_with_attr(
            "_neoscore_time_signature_type_marker"
        )

    def active_key_signature_at(self, pos_x: Unit) -> Optional[KeySignature]:
        """Return the active key signature at a given x position, if any."""
        return next(
            (sig for (sig_x, sig) in reversed(self.key_signatures) if sig_x <= pos_x),
            None,
        )

    def middle_c_at(self, pos_x: Unit) -> Unit:
        """Find the y-axis staff position of middle-c at a given point.

        Looks for clefs and other transposing modifiers to determine
        the position of middle-c.

        Raises:
            NoClefError:
                If no clef is active at the given position.
        """
        clef = self.active_clef_at(pos_x)
        if clef is None:
            raise NoClefError
        return clef.middle_c_staff_position

    def y_on_ledger(self, pos_y: Unit) -> bool:
        """Determine if a y-axis position is approximately at a ledger line position

        This is true for any whole-number staff position outside the staff
        """
        return (not self.y_inside_staff(pos_y)) and self.unit(
            pos_y
        ).display_value % 1 == 0

    def ledgers_needed_for_y(self, position: Unit) -> List[Unit]:
        """Find the y positions of all ledgers needed for a given y position"""
        # Work on positions as integers for simplicity
        start = int(self.unit(position).display_value)
        if start < 0:
            return [self.unit(pos) for pos in range(start, 0, 1)]
        elif start > self.line_count - 1:
            return [self.unit(pos) for pos in range(start, self.line_count - 1, -1)]
        else:
            return []

    @staticmethod
    def _make_unit_class(staff_unit_size: Unit) -> Type[Unit]:
        """Create a Unit class with a ratio of 1 to a staff unit size."""
        return make_unit_class("StaffUnit", staff_unit_size.base_value)

    def register_layout_controllers(self):
        # This is known to have some limitations in some cases when staves in a group
        # have different key signatures. See issue #28.
        flowable = self.flowable
        if not flowable:
            return
        staff_flowable_x = flowable.descendant_pos_x(self)
        flowable.add_margin_controller(
            MarginController(
                staff_flowable_x, self.unit(StaffGroup.RIGHT_PADDING), "_neoscore_staff"
            )
        )
        for clef_x, clef in self.clefs:
            flowable_x = staff_flowable_x + clef_x
            margin_needed = clef.bounding_rect.width
            if margin_needed != ZERO:
                # Some clefs can be zero-width; don't add padding in that case.
                margin_needed += self.unit(StaffGroup.CLEF_LEFT_PADDING)
            flowable.add_margin_controller(
                MarginController(flowable_x, margin_needed, "_neoscore_clef")
            )
        # Assume that key signatures have the same width in all clefs
        for key_sig_x, key_sig in self.key_signatures:
            flowable_x = staff_flowable_x + key_sig_x
            flowable.add_margin_controller(
                MarginController(
                    flowable_x,
                    key_sig.visual_width + self.unit(StaffGroup.KEY_SIG_LEFT_PADDING),
                    "_neoscore_key_signature",
                )
            )
        for time_sig in self.descendants_with_attribute(
            "_neoscore_time_signature_type_marker"
        ):
            flowable_x = flowable.descendant_pos_x(time_sig)
            flowable.add_margin_controller(
                MarginController(
                    flowable_x,
                    time_sig.visual_width + self.unit(StaffGroup.TIME_SIG_LEFT_PADDING),
                    "_neoscore_time_signature",
                )
            )
            # Cancel the margin controller immediately after it, this way time
            # signatures only affect margins if they lie right around a line start.
            flowable.add_margin_controller(
                MarginController(flowable_x + Unit(1), ZERO, "_neoscore_time_signature")
            )

    def fringe_layout_for_isolated_staff(
        self, location: Optional[NewLine]
    ) -> StaffFringeLayout:
        if location:
            staff_pos_x = location.flowable_x - self.flowable.descendant_pos_x(self)
            if staff_pos_x < ZERO:
                # This happens on the first line of a staff positioned at x>0 relative
                # to its flowable.
                staff_pos_x = ZERO
        else:
            staff_pos_x = ZERO
        # Work right-to-left through different fringe layers
        current_x = -self.unit(StaffGroup.RIGHT_PADDING)
        clef = self.active_clef_at(staff_pos_x)
        key_sig = self.active_key_signature_at(staff_pos_x)
        time_sig = next(
            (sig for x, sig in self.time_signatures if x == staff_pos_x), None
        )
        clef_fringe_pos = current_x
        key_signature_fringe_pos = current_x
        time_signature_fringe_pos = current_x

        if time_sig:
            current_x -= time_sig.visual_width
            time_signature_fringe_pos = current_x
            current_x -= self.unit(StaffGroup.TIME_SIG_LEFT_PADDING)
        if key_sig:
            current_x -= key_sig.visual_width
            key_signature_fringe_pos = current_x
            current_x -= self.unit(StaffGroup.KEY_SIG_LEFT_PADDING)
        if clef:
            clef_width = clef.bounding_rect.width
            current_x -= clef_width
            clef_fringe_pos = current_x
            if clef_width != ZERO:
                # Some clefs can be zero-width (invisible) - don't pad in this case.
                current_x -= self.unit(StaffGroup.CLEF_LEFT_PADDING)
        staff_fringe_pos = current_x
        return StaffFringeLayout(
            staff_pos_x,
            staff_fringe_pos,
            clef_fringe_pos,
            key_signature_fringe_pos,
            time_signature_fringe_pos,
        )



from __future__ import annotations

from typing import Optional

from neoscore.core.brush import Brush, BrushDef
from neoscore.core.pen import Pen, PenDef
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject


class PaintedObject(PositionedObject):
    """A ``PositionedObject`` which is painted with a pen (outline) and brush (fill).

    This is mostly meant to be used as a superclass for other classes which need this
    behavior.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """
        Args:
            pos: The position of the object relative to its parent
            pen: The pen to draw outlines with.
            brush: The brush to fill shapes with.
            parent: The parent object or None
        """
        super().__init__(pos, parent)
        self.pen = pen
        self.brush = brush
        self._children: List[PositionedObject] = []

    @property
    def pen(self) -> Pen:
        """The pen to draw outlines with"""
        return self._pen

    @pen.setter
    def pen(self, value: PenDef):
        if value:
            self._pen = Pen.from_def(value)
        else:
            self._pen = Pen()

    @property
    def brush(self) -> Brush:
        """The brush to draw outlines with.

        As a convenience, this may be set with a hex color string for a solid color
        brush of that color. For brushes using alpha channels and non-solid-color fill
        patterns, a fully initialized brush must be passed to this.
        """
        return self._brush

    @brush.setter
    def brush(self, value: BrushDef):
        if value:
            self._brush = Brush.from_def(value)
        else:
            self._brush = Brush()



from typing import List, Optional, Union

from neoscore.core.layout_controllers import NewLine
from neoscore.core.music_font import MusicFont
from neoscore.core.music_path import MusicPath
from neoscore.core.pen import Pen, PenDef
from neoscore.core.point import ORIGIN, Point
from neoscore.core.units import ZERO, Unit
from neoscore.western.abstract_staff import AbstractStaff
from neoscore.western.multi_staff_object import MultiStaffObject
from neoscore.western.staff_group import StaffGroup


class SystemLine(MultiStaffObject, MusicPath):

    """A line connecting staves at the beginning of every system.

    This is drawn in the fringe of every staff system in the specified group.
    """

    def __init__(
        self,
        staves: Union[StaffGroup, List[AbstractStaff]],
        font: Optional[MusicFont] = None,
        pen: Optional[PenDef] = None,
    ):
        """
        Args:
            staves: The staves spanned. If a raw list of staves is given, it must be
                in descending order.
            font: If provided, this overrides the font in the parent (top) staff.
            pen: An override to the pen normally derived from the active music font.
        """
        MultiStaffObject.__init__(self, staves)
        font = font or self.highest.music_font
        if pen is None:
            pen = Pen(thickness=font.engraving_defaults["staffLineThickness"])
        MusicPath.__init__(self, ORIGIN, self.highest, font, pen=pen)
        self.move_to(ZERO, self.highest.barline_extent[0])
        self.line_to(ZERO, self.map_to(self.lowest).y + self.lowest.barline_extent[1])

    @property
    def breakable_length(self) -> Unit:
        """This class's breakable length is that of its highest staff"""
        return self.highest.breakable_length

    def _render_occurrence(
        self,
        pos: Point,
        flowable_line: Optional[NewLine],
    ):
        fringe_layout = self.highest.fringe_layout_at(flowable_line)
        super().render_complete(
            Point(pos.x + fringe_layout.staff, pos.y), flowable_line
        )

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        self._render_occurrence(pos, flowable_line)

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        self._render_occurrence(pos, flowable_line)

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        self._render_occurrence(pos, flowable_line)

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        self._render_occurrence(pos, flowable_line)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, NamedTuple, Optional, Union

from PyQt5.QtGui import QPainterPath
from typing_extensions import TypeAlias

from neoscore.core.units import Unit
from neoscore.interface.brush_interface import BrushInterface
from neoscore.interface.pen_interface import PenInterface
from neoscore.interface.positioned_object_interface import PositionedObjectInterface
from neoscore.interface.qt.converters import point_to_qt_point_f
from neoscore.interface.qt.q_clipping_path import QClippingPath


class ResolvedMoveTo(NamedTuple):
    """A canvas-space move-to element."""

    # Add blank docstrings to suppress ugly default namedtuple docstring used by Sphinx
    x: Unit
    """"""
    y: Unit
    """"""


class ResolvedLineTo(NamedTuple):
    """A canvas-space line-to element."""

    x: Unit
    """"""
    y: Unit
    """"""


class ResolvedCurveTo(NamedTuple):
    """A canvas-space curve-to element."""

    c1_x: Unit
    """"""
    c1_y: Unit
    """"""
    c2_x: Unit
    """"""
    c2_y: Unit
    """"""
    end_x: Unit
    """"""
    end_y: Unit
    """"""


ResolvedPathElement: TypeAlias = Union[ResolvedMoveTo, ResolvedLineTo, ResolvedCurveTo]
"""A path element whose position is relative to its path"""


@dataclass(frozen=True)
class PathInterface(PositionedObjectInterface):
    """Interface for a generic graphic path object."""

    brush: BrushInterface

    pen: PenInterface

    elements: List[ResolvedPathElement]

    background_brush: Optional[BrushInterface] = None

    clip_start_x: Optional[Unit] = None
    """The local starting position of the drawn region in the glyph.

    Use ``None`` to render from the start
    """

    clip_width: Optional[Unit] = None
    """The width of the visible region.

    Use ``None`` to render to the end.
    """

    @staticmethod
    def create_qt_path(elements: List[ResolvedPathElement]) -> QPainterPath:
        path = QPainterPath()
        path.setFillRule(1)

        for el in elements:
            if isinstance(el, ResolvedLineTo):
                path.lineTo(el.x.base_value, el.y.base_value)
            elif isinstance(el, ResolvedMoveTo):
                path.moveTo(el.x.base_value, el.y.base_value)
            elif isinstance(el, ResolvedCurveTo):
                path.cubicTo(
                    el.c1_x.base_value,
                    el.c1_y.base_value,
                    el.c2_x.base_value,
                    el.c2_y.base_value,
                    el.end_x.base_value,
                    el.end_y.base_value,
                )
            else:
                raise TypeError("Unknown ResolvedPathElement type")
        return path

    def render(self):
        """Render the path to the scene."""
        self._register_qt_object(self._create_qt_object())

    def _create_qt_object(self) -> QClippingPath:
        painter_path = PathInterface.create_qt_path(self.elements)
        qt_object = QClippingPath(
            painter_path,
            self.clip_start_x.base_value if self.clip_start_x is not None else 0,
            self.clip_width.base_value if self.clip_width is not None else None,
            self.scale,
            self.rotation,
            self.background_brush.qt_object if self.background_brush else None,
            defer_geometry_calculation=True,
            transform_origin=point_to_qt_point_f(self.transform_origin),
        )
        qt_object.setPos(point_to_qt_point_f(self.pos))
        qt_object.setBrush(self.brush.qt_object)
        qt_object.setPen(self.pen.qt_object)
        qt_object.update_geometry()
        return qt_object



from __future__ import annotations

from typing import Any, Optional, Union

from typing_extensions import TypeAlias

from neoscore.core.color import Color, ColorDef
from neoscore.core.pen_cap_style import PenCapStyle
from neoscore.core.pen_join_style import PenJoinStyle
from neoscore.core.pen_pattern import PenPattern
from neoscore.core.units import ZERO, Unit
from neoscore.interface.pen_interface import PenInterface


class Pen:

    """A pen describing how shape outlines are drawn."""

    _default_color = Color("#000000")

    def __init__(
        self,
        color: Optional[ColorDef] = None,
        thickness: Optional[Unit] = ZERO,
        pattern: PenPattern = PenPattern.SOLID,
        join_style: PenJoinStyle = PenJoinStyle.MITER,
        cap_style: PenCapStyle = PenCapStyle.FLAT,
    ):
        """
        Args:
            color: The stroke color. Defaults to black unless changed globally by
                :obj:`.neoscore.set_default_color`.
            thickness: The stroke thickness. A value of ``ZERO`` (the default) indicates
                a display pixel width.
            pattern: The stroke pattern. Defaults to a solid line.
            join_style: The appearance of line joints. Defaults to a bevel join
            cap_style: The appearance of line ends. Defaults to a flat cap
        """
        if color is None:
            self._color = Pen._default_color
        else:
            self._color = Color.from_def(color)
        self._thickness = thickness
        self._pattern = pattern
        self._join_style = join_style
        self._cap_style = cap_style
        self._regenerate_interface()

    @classmethod
    def from_existing(
        cls,
        pen: Pen,
        color: Optional[ColorDef] = None,
        thickness: Optional[Unit] = None,
        pattern: Optional[PenPattern] = None,
        join_style: Optional[PenJoinStyle] = None,
        cap_style: Optional[PenCapStyle] = None,
    ) -> Pen:
        """Derive a pen from another, overriding any provided fields."""
        return cls(
            color or pen.color,
            thickness or pen.thickness,
            pattern or pen.pattern,
            join_style or pen.join_style,
            cap_style or pen.cap_style,
        )

    @classmethod
    def no_pen(cls) -> Pen:
        """Create a non-drawing pen."""
        return Pen(pattern=PenPattern.INVISIBLE)

    @classmethod
    def from_def(cls, pen_def: PenDef) -> Pen:
        if isinstance(pen_def, Pen):
            return pen_def
        else:
            return Pen(pen_def)

    def _regenerate_interface(self):
        self._interface = PenInterface(
            self.color,
            self.thickness,
            self.pattern,
            self.join_style,
            self.cap_style,
        )

    @property
    def color(self) -> Color:
        """The color for the pen.

        This can be set with a :obj:`.ColorDef` shorthand."""
        return self._color

    @color.setter
    def color(self, value: ColorDef):
        self._color = Color.from_def(value)
        self._regenerate_interface()

    @property
    def thickness(self) -> Unit:
        """The pen stroke thickness."""
        return self._thickness

    @thickness.setter
    def thickness(self, value: Unit):
        self._thickness = value
        self._regenerate_interface()

    @property
    def pattern(self) -> PenPattern:
        """The stroke pattern."""
        return self._pattern

    @pattern.setter
    def pattern(self, value: PenPattern):
        self._pattern = value
        self._regenerate_interface()

    @property
    def join_style(self) -> PenJoinStyle:
        """The style of line sharp line joins.

        This style has no effect on curved paths.
        """
        return self._join_style

    @join_style.setter
    def join_style(self, value: PenJoinStyle):
        self._join_style = value
        self._regenerate_interface()

    @property
    def cap_style(self) -> PenCapStyle:
        """the style of unclosed path caps with this pen.

        This style has no effect on closed paths."""
        return self._cap_style

    @cap_style.setter
    def cap_style(self, value: PenCapStyle):
        self._cap_style = value
        self._regenerate_interface()

    @property
    def interface(self) -> PenInterface:
        return self._interface

    def __eq__(self, other: Any) -> bool:
        """Pens are compared by their attributes"""
        return (
            isinstance(other, Pen)
            and self.color == other.color
            and self.thickness == other.thickness
            and self.pattern == other.pattern
            and self.join_style == other.join_style
            and self.cap_style == other.cap_style
        )


PenDef: TypeAlias = Union[Pen, str]
"""A ``Pen`` or a color hex string to be passed to an otherwise default ``Pen``."""



from __future__ import annotations

from typing import List, Optional

from neoscore.core import neoscore
from neoscore.core.brush import Brush
from neoscore.core.directions import DirectionY
from neoscore.core.has_music_font import HasMusicFont
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.path import Path
from neoscore.core.pen import Pen
from neoscore.core.point import ORIGIN, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.spanner_2d import Spanner2D
from neoscore.core.text_alignment import AlignmentX, AlignmentY
from neoscore.core.units import ZERO


class Tuplet(Spanner2D, PositionedObject, HasMusicFont):
    """A polyrhythm indicator such as triplet or 3:4.

    This tuplet indicator spans a group of notes labelling them
    as triplet or more complex polyrhythms such as 5:4.

    At the centre of this line is the tuplet number.
    More complex ratios can be declared
    such as 7:8, 11:16, 12:14.

    Optional parameters enable bracket visibility,
    and bracket direction.

    """

    def __init__(
        self,
        start: PointDef,
        start_parent: PositionedObject,
        end: PointDef,
        end_parent: Optional[PositionedObject] = None,
        indicator_text: str = "3",
        include_bracket: bool = True,
        bracket_dir: DirectionY = DirectionY.DOWN,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            start: The starting point.
            start_parent: The parent for the starting position. If no font is given,
                this or one of its ancestors must implement :obj:`.HasMusicFont`.
            end: The spanner end point position. This must be to the right of the start position
            end_parent: An object either in a Staff or
                a staff itself. The root staff of this *must* be the same
                as the root staff of ``start_parent``. If omitted, the
                stop point is relative to the start point.
            indicator_text: The tuplet indicator text drawn at the middle of the spanner.
                This should contain only of numbers and colons, for example "3" or "5:4".
                Any other character will cause a ``ValueError`` to be raised.
            include_bracket: Whether to draw a bracket spanning the tuplet
            bracket_dir: The direction the line's ending hook points.
            font: If provided, this overrides any font found in the ancestor chain.
        """
        PositionedObject.__init__(self, start, start_parent)
        Spanner2D.__init__(self, end, end_parent or self)
        if font is None:
            font = HasMusicFont.find_music_font(start_parent)
        self._music_font = font
        self.direction = bracket_dir

        # Calculate the bracket end length
        self.bracket_height = self.music_font.unit(self.direction.value)

        # Create bracket path if needed
        self.bracket: Optional[Path] = (
            self._create_bracket() if include_bracket else None
        )

        # Convert ratio text to SMuFL glyphs
        self.smufl_text = self._indicator_to_glyph_names(indicator_text)

        # Create line text object; will paint over bracket line
        spanner_center = self.point_along_spanner(0.5)
        self.indicator = MusicText(
            (spanner_center.x, spanner_center.y + self.bracket_height),
            self,
            self.smufl_text,
            font,
            background_brush=neoscore.background_brush,
            alignment_x=AlignmentX.CENTER,
            alignment_y=AlignmentY.CENTER,
        )

    def _create_bracket(self) -> Path:
        bracket = Path(
            ORIGIN,
            self,
            Brush.no_brush(),
            Pen(thickness=self.music_font.engraving_defaults["tupletBracketThickness"]),
        )
        # Draw opening crook
        bracket.line_to(ZERO, self.bracket_height)
        # Draw spanning line
        bracket.line_to(
            self.end_pos.x, self.end_y + self.bracket_height, self.end_parent
        )
        # Draw end crook
        bracket.line_to(self.end_pos.x, self.end_y, self.end_parent)
        return bracket

    @staticmethod
    def _indicator_to_glyph_names(indicator: str) -> List[str]:
        """Converts indicator text to corresponding SMuFL names

        Args:
            number: the original ratio string
        """
        smufl_list = []
        for char in indicator:
            if char == ":":
                smufl_list.append("tupletColon")
            else:
                smufl_list.append(f"tuplet{char}")
        return smufl_list

    @property
    def music_font(self) -> MusicFont:
        return self._music_font



import sys
from enum import Enum, auto


class PlatformType(Enum):
    """Common operating systems"""

    BSD = auto()
    LINUX = auto()
    MAC = auto()
    WINDOWS = auto()
    UNKNOWN = auto()


def current_platform() -> PlatformType:
    """Get the type of the running platform"""
    if sys.platform.startswith("freebsd"):
        return PlatformType.BSD
    elif sys.platform.startswith("linux"):
        return PlatformType.LINUX
    elif sys.platform == "darwin":
        return PlatformType.MAC
    elif sys.platform == "win32" or sys.platform == "cygwin":
        return PlatformType.WINDOWS
    return PlatformType.unkown



from __future__ import annotations

from typing import NamedTuple, Tuple, Union, cast

from typing_extensions import TypeAlias

from neoscore.core.units import ZERO, Unit


class Point(NamedTuple):
    """A two-dimensional point.

    The x-axis grows left-to right, and the y-axis grows top-to-bottom.
    """

    x: Unit
    """The horizontal X-axis value"""

    y: Unit
    """The vertical Y-axis value"""

    @staticmethod
    def from_def(point_def: PointDef) -> Point:
        if hasattr(point_def, "x"):
            return cast(Point, point_def)
        return Point(*point_def)

    def __add__(self, other: Point) -> Point:
        """Points are added by adding their x and y values respectively"""
        try:
            return Point(self.x + other.x, self.y + other.y)
        except AttributeError:
            raise TypeError

    def __sub__(self, other: Point) -> Point:
        """Points are subtracted by subtracting their x and y values respectively"""
        try:
            return Point(self.x - other.x, self.y - other.y)
        except AttributeError:
            raise TypeError

    def __mul__(self, other: float) -> Point:
        """Points may be multiplied by non-unit scalars.

        This is done by multiplying the x and y values by that scalar.
        """
        try:
            return Point(self.x * other, self.y * other)
        except AttributeError:
            raise TypeError

    def __abs__(self) -> Point:
        """Get a Point whose x and y values are the absolute values of this point's."""
        return Point(abs(self.x), abs(self.y))

    def __neg__(self) -> Point:
        """Get a Point whose x and y values are the negation of this point's."""
        return Point(-self.x, -self.y)


ORIGIN = Point(ZERO, ZERO)
"""Shorthand for a point at ``ZERO, ZERO``"""

PointDef: TypeAlias = Union[Point, Tuple[Unit, Unit]]
"""A Point or an argument tuple for one"""



from dataclasses import dataclass
from typing import Dict, NamedTuple, Optional

from PyQt5.QtGui import QFont, QPainterPath

from neoscore.core.units import Unit
from neoscore.interface.brush_interface import BrushInterface
from neoscore.interface.font_interface import FontInterface
from neoscore.interface.pen_interface import PenInterface
from neoscore.interface.positioned_object_interface import PositionedObjectInterface
from neoscore.interface.qt.converters import point_to_qt_point_f
from neoscore.interface.qt.q_clipping_path import QClippingPath


class _CachedTextKey(NamedTuple):
    text: str
    family_name: str
    weight: Optional[int]
    italic: bool


class _CachedTextPath(NamedTuple):
    path: QPainterPath
    generation_font_size: int


_PATH_CACHE: Dict[_CachedTextKey, _CachedTextPath] = {}

"""NOTE: We can actually optimize this even further. We can modify
q_clipping_path so it explicitly stores paint results in the global
QPixmapCache. If this were specialized to just text items, the cache
key would be like _CachedTextKey, except it also includes font size
and scale. This would allow us to not only cache the paths being sent
to Qt objects, but the rendered pixmaps themselves. This would be very
efficient for us because so many constructed path objects are
identical. Furthermore, if I manage to move the runtime to the OpenGL
system, I believe these rendered pixmaps would reside directly in GPU
memory.

see https://doc.qt.io/qt-5/qgraphicsitem.html#setCacheMode
"""


@dataclass(frozen=True)
class TextInterface(PositionedObjectInterface):

    """An interface for graphical text objects."""

    brush: BrushInterface

    pen: PenInterface

    text: str

    font: FontInterface

    background_brush: Optional[BrushInterface] = None

    clip_start_x: Optional[Unit] = None
    """The local starting position of the drawn region in the glyph.

    Use ``None`` to render from the start
    """

    clip_width: Optional[Unit] = None
    """The width of the visible region.

    Use ``None`` to render to the end.
    """

    def render(self):
        """Render the line to the scene."""
        self._register_qt_object(self._create_qt_object())

    def _create_qt_object(self) -> QClippingPath:
        """Create and return this interface's underlying Qt object"""
        qt_object = self._get_path(self.text, self.font, self.scale)
        qt_object.setPos(point_to_qt_point_f(self.pos))
        qt_object.setBrush(self.brush.qt_object)
        qt_object.setPen(self.pen.qt_object)
        qt_object.update_geometry()
        return qt_object

    def _get_path(self, text: str, font: FontInterface, scale: float) -> QClippingPath:
        qt_font = font.qt_object
        needed_font_size = qt_font.pixelSize()
        key = _CachedTextKey(text, font.family_name, font.weight, font.italic)
        cached_result = _PATH_CACHE.get(key)
        if cached_result:
            cache_scale = needed_font_size / cached_result.generation_font_size
            scale *= cache_scale
            path = cached_result.path
        else:
            path = TextInterface._create_qt_path(text, qt_font)
            _PATH_CACHE[key] = _CachedTextPath(path, needed_font_size)
        return QClippingPath(
            path,
            self.clip_start_x.base_value if self.clip_start_x is not None else 0,
            self.clip_width.base_value if self.clip_width is not None else None,
            scale,
            self.rotation,
            self.background_brush.qt_object if self.background_brush else None,
            defer_geometry_calculation=True,
            transform_origin=point_to_qt_point_f(self.transform_origin),
        )

    @staticmethod
    def _create_qt_path(text: str, font: QFont) -> QPainterPath:
        qt_path = QPainterPath()
        qt_path.addText(0, 0, font, text)
        qt_path.setFillRule(1)
        return qt_path



from __future__ import annotations

import math
from dataclasses import dataclass

from typing_extensions import TypeAlias

from neoscore.core.math_helpers import is_power_of_2

BaseDuration: TypeAlias = int
"""A base duration used to set notehead and rest glyphs.

Must be 0 or a power of 2. The values correspond to durations like so:

* 0: Double whole note
* 1: Whole note
* 2: Half note
* 4: Quarter note (etc)
"""


@dataclass(frozen=True)
class DurationDisplay:

    """The basic information about a :obj:`.Duration` needed to write a note with."""

    base_duration: BaseDuration

    dot_count: int

    def __post_init__(self):
        if not (is_power_of_2(self.base_duration) or self.base_duration == 0):
            raise ValueError("base_duration must be 0 or a power of 2")

    @property
    def flag_count(self) -> int:
        """Flag and beam count needed if used in a note"""
        if self.base_duration <= 4:
            return 0
        return int(math.log2(self.base_duration) - 2)

    @property
    def requires_stem(self) -> bool:
        return self.base_duration > 1



from __future__ import annotations

from typing import Optional

from neoscore.core.brush import Brush
from neoscore.core.music_font import MusicFont
from neoscore.core.music_path import MusicPath
from neoscore.core.pen import Pen
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.spanner import Spanner
from neoscore.core.units import Unit


class PedalLine(Spanner, MusicPath):

    """A line-style pedal marking with optional half-lift (^) marks.

    The mark begins with an upward crook, continues as a horizontal
    line interrupted with optional half-lift upward triangular marks,
    and terminates with another upward crook.

    The object's position is that of the top of the opening crook;
    this y-axis position is also the height two which the half-lift
    marks extend. The main line of the path is drawn 1 staff unit
    below this position.

    A list of half-lift x-axis positions may be provided as either
    simple x-axis Unit values (relative to the line's start position)
    or as 2-tuples of x-axis positions and position parents:

    >>> PedalLine(start=(Mm(10), staff.unit(8)), start_parent=staff,
    ...           end_x=Mm(200), end_parent=None,
    ...           half_lift_positions=[
    ...               Mm(20),               # Relative to start_parent
    ...               Mm(40),               # Relative to start_parent
    ...               (Mm(1), some_chord)]  # 1 mm right of some_chord
    ...          ) # doctest: +SKIP

    While this is a path, it requires a music font from which to
    derive its appearance.
    """

    def __init__(
        self,
        start: PointDef,
        start_parent: PositionedObject,
        end_x: Unit,
        end_parent: Optional[PositionedObject] = None,
        half_lift_positions: List[Unit | Tuple[Unit, PositionedObject]] = None,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            start: The starting position of the pedal line.
            start_parent: The parent for the starting position. If no font is given,
                this or one of its ancestors must implement ``HasMusicFont``.
            end_x: x position of the end point relative to end_parent
            end_parent: parent for end_x
            half_lift_positions: A list of x-axis positions along the line
                where half lift notches should be drawn.
            font: If provided, this overrides any font found in the ancestor chain.
        """
        MusicPath.__init__(self, start, start_parent, font, Brush.no_brush(), Pen())
        self.pen.thickness = self.music_font.engraving_defaults["pedalLineThickness"]
        Spanner.__init__(self, end_x, end_parent or self)
        self.half_lift_positions = half_lift_positions
        self._draw_path()

    def _draw_path(self):
        """Draw the path according to this object's attributes.

        Returns: None
        """
        # Set up constants
        # The vertical extent of the shape
        descent = self.music_font.unit(1)
        # The horizontal distance extended to from half lift positions
        half_lift_x_extent = self.music_font.unit(0.5)

        # Draw opening crook
        self.line_to(self.music_font.unit(0), descent)

        # Draw half lift positions, if any exist
        if self.half_lift_positions:
            for pos in self.half_lift_positions:
                if pos >= self.end_x - (half_lift_x_extent * 2):
                    raise AttributeError("Invalid half lift position: {}".format(pos))
                if isinstance(pos, tuple):
                    lift_parent = pos[1]
                    lift_center_x = pos[0]
                    lift_center_y = lift_parent.map_to(self).y + descent
                elif isinstance(pos, Unit):
                    lift_parent = self
                    lift_center_x = pos
                    lift_center_y = descent
                else:
                    raise AttributeError("Invalid half lift position: {}".format(pos))
                lift_left_x = lift_center_x - half_lift_x_extent
                lift_right_x = lift_center_x + half_lift_x_extent
                self.line_to(lift_left_x, lift_center_y, lift_parent)
                self.line_to(lift_center_x, lift_center_y - descent, lift_parent)
                self.line_to(lift_right_x, lift_center_y, lift_parent)

        # Draw closing crook
        self.line_to(self.end_x, self.end_y + descent, self.end_parent)
        self.line_to(self.end_x, self.end_y, self.end_parent)



from __future__ import annotations

from functools import reduce
from typing import Optional

from neoscore.core.brush import BrushDef
from neoscore.core.directions import DirectionY
from neoscore.core.music_font import MusicFont
from neoscore.core.pen import PenDef
from neoscore.core.point import Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.rect import Rect
from neoscore.core.repeating_music_text_line import RepeatingMusicTextLine
from neoscore.western.chordrest import Chordrest


class ArpeggioLine(RepeatingMusicTextLine):

    """An arpeggio roll line, optionally with an arrow.

    This is a thin convenience wrapper around ``RepeatingMusicTextLine``.
    """

    _MAIN_GLYPH = "wiggleArpeggiatoUp"
    _ARROW_GLYPH = "wiggleArpeggiatoUpArrow"

    def __init__(
        self,
        start: PointDef,
        start_parent: Optional[PositionedObject],
        end_pos: PointDef,
        end_parent: Optional[PositionedObject],
        include_arrow: bool = False,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        background_brush: Optional[BrushDef] = None,
    ):
        """
        Args:
            start: The starting point.
            start_parent: If no font is given, this or one of its ancestors
                must implement :obj:`.HasMusicFont`.
            end_pos: The stopping point.
            end_parent: The parent for the ending position.
                If ``None``, defaults to ``self``.
            include_arrow: Whether to end the line with an arrow. This will be
                attached to the end position pointing away from the starting position.
            font: If provided, this overrides any font found in the ancestor chain.
            brush: The brush to fill in text shapes with.
            pen: The pen to trace text outlines with. This defaults to no pen.
            background_brush: Optional brush used to paint the text's bounding rect
                behind it.
        """
        end_cap_text = ArpeggioLine._ARROW_GLYPH if include_arrow else None
        super().__init__(
            start,
            start_parent,
            end_pos,
            end_parent,
            ArpeggioLine._MAIN_GLYPH,
            None,
            end_cap_text,
            font,
            brush,
            pen,
            background_brush,
        )

    @classmethod
    def for_chord(
        cls,
        chordrest: Chordrest,
        arrow_direction: Optional[DirectionY] = None,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
        background_brush: Optional[BrushDef] = None,
    ) -> ArpeggioLine:
        """Convenience constructor for an arpeggio line attached to a chord.

        Args:
            chordrest: The chord to attach the line to. This must not be a rest.
            arrow_direction: If provided, cap the line with an arrow pointing this way.
            font: A font override. If omitted, the chord's font is used.
            brush: The brush to fill in text shapes with.
            pen: The pen to trace text outlines with. This defaults to no pen.
            background_brush: Optional brush used to paint the text's bounding rect
                behind it.
        """
        if font:
            unit = font.unit
        else:
            unit = chordrest.staff.unit
        # Merge the bounding rects of all accidentals and noteheads
        # (offset relative to chordrest)
        guide_rect = reduce(
            lambda a, b: a.merge(b),
            [n.bounding_rect.offset(n.pos) for n in chordrest.noteheads],
        )
        if chordrest.accidentals:
            accidental_rect = reduce(
                lambda a, b: a.merge(b),
                [a.bounding_rect.offset(a.pos) for a in chordrest.accidentals],
            )
            guide_rect = Rect(
                accidental_rect.x, guide_rect.y, guide_rect.width, guide_rect.height
            )
        top = Point(guide_rect.x - unit(0.5), guide_rect.y - unit(0.25))
        bottom = Point(top.x, top.y + guide_rect.height + unit(0.5))
        include_arrow = arrow_direction is not None
        if arrow_direction == DirectionY.UP:
            start = bottom
            end = top
        else:
            start = top
            end = bottom
        return ArpeggioLine(
            start,
            chordrest,
            end,
            chordrest,
            include_arrow,
            font,
            brush,
            pen,
            background_brush,
        )



from typing import List, Union

from neoscore.core.units import Unit
from neoscore.western.abstract_staff import AbstractStaff
from neoscore.western.staff_group import StaffGroup


class MultiStaffObject:

    """An object which spans several staves.

    This is a mixin class for :obj:`.PositionedObject` classes. Such
    ``PositionedObject``\ s must have their highest staff also be their parent such that
    ``self.parent == self.highest``.
    """

    def __init__(self, staves: Union[StaffGroup, List[AbstractStaff]]):
        """
        Args:
            staves: The staves this is associated with. If a raw list of staves
                is given, it must be in visually descending order.
        """
        if isinstance(staves, StaffGroup):
            self._staves = staves.staves
        else:
            self._staves = staves

    @property
    def staves(self) -> List[AbstractStaff]:
        """The staves this is associated with, given in descending order."""
        return self._staves

    @staves.setter
    def staves(self, value: List[AbstractStaff]):
        self._staves = value

    @property
    def highest(self) -> AbstractStaff:
        """Shorthand for ``staves[0]``"""
        return self.staves[0]

    @property
    def lowest(self) -> AbstractStaff:
        """Shorthand for ``staves[-1]``"""
        return self.staves[-1]

    @property
    def vertical_span(self) -> Unit:
        """The vertical distance covered by the staves.

        This distance extends from the top line of the top staff to
        the bottom line of the bottom staff.
        """
        highest = self.staves[0]
        lowest = self.staves[-1]
        return highest.unit(highest.map_to(lowest).y + lowest.height)

    @property
    def center_y(self) -> Unit:
        """The vertical center of the staves spanned.

        This value is relative to the top of the highest staff.
        """
        return self.vertical_span / 2



import re

from neoscore.core.directions import DirectionY
from neoscore.core.exceptions import InvalidIntervalError


class Interval:
    """An abstract pitch interval."""

    _shorthand_regex = re.compile("^([ad])([mMPdA])([1-9]\d*)$")
    _base_pc_deltas = {1: 0, 2: 2, 3: 4, 4: 5, 5: 7, 6: 9, 7: 11}
    _qualities_in_english = {
        "m": "minor",
        "M": "Major",
        "d": "diminished",
        "A": "Augmented",
    }
    _perfectable_distances = [1, 4, 5]

    def __init__(self, specifier: str):
        """
        The interval specifier should be a string in the form:
        ``[direction][quality][distance]`` where:

        * ``direction`` is one of:

            * ``'a'`` for ascending
            * ``'d'`` for descending

        * ``quality`` is one of:

            * ``'m'`` for minor
            * ``'M'`` for Major
            * ``'P'`` for Perfect
            * ``'d'`` for diminished
            * ``'A'`` for Augmented

        * ``distance`` is any positive integer indicating the
          interval distance.

        Some examples:

            * ``Interval('aM3')`` signifies an ascending major third
            * ``Interval('dA9')`` signifies a descending augmented ninth
        """
        match = Interval._shorthand_regex.match(specifier)
        if match is None:
            raise InvalidIntervalError
        self._direction = DirectionY.UP if match.group(1) == "a" else DirectionY.DOWN
        self._quality = match.group(2)
        self._distance = int(match.group(3))
        # Check against invalid edge case intervals
        if (
            self.simple_distance in Interval._perfectable_distances
            and self._quality
            not in [
                "P",
                "d",
                "A",
            ]
        ):
            # unisons, fourths, fifths, and their compounds
            # can only be perfect, augmented, or diminished
            raise InvalidIntervalError

    def __repr__(self):
        direction_str = "a" if self.direction == DirectionY.UP else "d"
        return f"Interval('{direction_str}{self.quality}{self.distance}')"

    def __hash__(self):
        return hash(self.direction) ^ hash(self.quality) ^ hash(self.distance)

    def __eq__(self, other):
        return (
            isinstance(other, Interval)
            and self.direction == other.direction
            and self.quality == other.quality
            and self.distance == other.distance
        )

    @property
    def direction(self) -> DirectionY:
        return self._direction

    @property
    def quality(self) -> str:
        return self._quality

    @property
    def quality_in_english(self) -> str:
        return Interval._qualities_in_english[self._quality]

    @property
    def distance(self) -> int:
        return self._distance

    @property
    def staff_distance(self) -> float:
        """The vertical offset needed to show this interval on a staff.

        The value is given in pseudo-staff-units.

        >>> Interval('aM3').staff_distance
        -1.0
        >>> Interval('dm3').staff_distance
        1.0
        >>> Interval('aP1').staff_distance
        0.0
        >>> Interval('dP8').staff_distance
        3.5
        """
        return ((self.distance - 1) * self.direction.value) / 2

    @property
    def simple_distance(self) -> float:
        """The simplified interval distance collapsing compound intervals.

        This value is the simplified version of ``self.distance`` where intervals
        larger than an octave are moved to be within one octave.

        Examples:
            >>> Interval('aM10').simple_distance
            3
            >>> Interval('dP12').simple_distance
            5
        """
        return ((self.distance - 1) % 7) + 1



from typing import Optional

from neoscore.core import neoscore
from neoscore.core.brush import BrushDef
from neoscore.core.font import Font
from neoscore.core.layout_controllers import MarginController, NewLine
from neoscore.core.painted_object import PaintedObject
from neoscore.core.pen import PenDef
from neoscore.core.point import ORIGIN, Point, PointDef
from neoscore.core.text import Text
from neoscore.core.text_alignment import AlignmentX, AlignmentY
from neoscore.core.units import Unit
from neoscore.western.abstract_staff import AbstractStaff
from neoscore.western.staff_object import StaffObject


class InstrumentName(PaintedObject, StaffObject):

    """A simple instrument name written in staff fringes

    When created, this causes a given string to be printed in the fringe of each staff
    system. The given position is relative to the top left corner of the staff in the
    fringe, and the text is right-aligned and vertically centered at this point.

    Instrument name changes within a staff are not currently supported.
    """

    def __init__(
        self,
        relative_fringe_pos: PointDef,
        staff: AbstractStaff,
        first_line_text: str,
        later_lines_text: Optional[str] = None,
        font: Optional[Font] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """
        Args:
            relative_fringe_pos: The position of the text in the fringe. This is
                relative to the top left corner of the staff in the fringe, and the
                text is right-aligned and vertically centered at this point.
            staff: The parent staff
            first_line_text: The text to write in the first line's fringe
            later_lines_text: Optional different text to write in lines after the first.
                This is useful for writing short-form instrument names. If ``None``, the
                first line text is used everywhere. Set to an empty string to make
                later lines blank.
            font: The font used. Defaults to the document-wide default font.
            brush: The brush to fill in text shapes with.
            pen: The pen to trace text outlines with. This defaults to no pen.
        """
        StaffObject.__init__(self, staff)
        PaintedObject.__init__(self, ORIGIN, staff, brush, pen)
        if font:
            self._font = font
        else:
            self._font = neoscore.default_font
        self.relative_fringe_pos = relative_fringe_pos
        self._first_line_text = first_line_text
        self._later_lines_text = later_lines_text

    @property
    def relative_fringe_pos(self) -> Point:
        return self._relative_fringe_pos

    @relative_fringe_pos.setter
    def relative_fringe_pos(self, value: PointDef):
        self._relative_fringe_pos = Point.from_def(value)

    @property
    def font(self) -> Font:
        """The text font"""
        return self._font

    @font.setter
    def font(self, value: Font):
        self._font = value

    @property
    def first_line_text(self) -> str:
        """The text to write in the first line's fringe"""
        return self._first_line_text

    @first_line_text.setter
    def first_line_text(self, value: str):
        self._first_line_text = value

    @property
    def later_lines_text(self) -> Optional[str]:
        """Optional different text to write in lines after the first.

        This is useful for writing short-form instrument names. If ``None``, the first
        line text is used everywhere. Set to an empty string to make later lines blank.
        """
        return self._later_lines_text

    @later_lines_text.setter
    def later_lines_text(self, value: Optional[str]):
        self._later_lines_text = value

    @property
    def _resolved_later_lines_text(self) -> str:
        if self.later_lines_text is None:
            return self.first_line_text
        else:
            return self.later_lines_text

    @property
    def breakable_length(self) -> Unit:
        """This class's breakable length is that of its staff"""
        return self.staff.breakable_length

    @property
    def first_line_visual_width(self) -> Unit:
        """The width of the first line of text"""
        return self.font.bounding_rect_of(self.first_line_text).width

    @property
    def later_lines_visual_width(self) -> Unit:
        """The width of later lines of text"""
        return self.font.bounding_rect_of(self._resolved_later_lines_text).width

    def _render_occurrence(
        self,
        render_call_pos: Point,
        flowable_line: Optional[NewLine],
        is_first_line: bool,
    ):
        fringe_layout = self.staff.fringe_layout_at(flowable_line)
        pos = Point(
            render_call_pos.x + self.relative_fringe_pos.x + fringe_layout.staff,
            render_call_pos.y + self.relative_fringe_pos.y,
        )
        if is_first_line:
            text = self.first_line_text
        else:
            text = self._resolved_later_lines_text
        if text == "":
            # Skip rendering blank strings
            return
        line_text = Text(
            pos,
            None,
            text,
            self.font,
            self.brush,
            self.pen,
            alignment_x=AlignmentX.RIGHT,
            alignment_y=AlignmentY.CENTER,
        )
        line_text.render()
        line_text.remove()

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        self._render_occurrence(pos, flowable_line, True)

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        self._render_occurrence(pos, flowable_line, True)

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        self._render_occurrence(pos, flowable_line, False)

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        self._render_occurrence(pos, flowable_line, False)

    def _register_layout_controllers(self):
        # This is known to have some limitations in some cases when a staff is added
        # later in a group than others and has a visually shorted instrument name than
        # the others. See issue #28.
        flowable = self.flowable
        if not flowable:
            return
        staff_flowable_x = flowable.descendant_pos_x(self.staff)
        flowable.add_margin_controller(
            MarginController(
                staff_flowable_x,
                self.first_line_visual_width,
                "_neoscore_instrument_name",
            )
        )
        # Update immediately after for later lines width
        flowable.add_margin_controller(
            MarginController(
                staff_flowable_x + Unit(1),
                self.later_lines_visual_width,
                "_neoscore_instrument_name",
            )
        )

    def pre_render_hook(self):
        super().pre_render_hook()
        self._register_layout_controllers()



from neoscore.core import neoscore
from neoscore.core.break_hint import BreakHint
from neoscore.core.brush import Brush
from neoscore.core.brush_pattern import BrushPattern
from neoscore.core.color import Color
from neoscore.core.directions import DirectionX, DirectionY
from neoscore.core.document import Document
from neoscore.core.flowable import Flowable
from neoscore.core.font import Font
from neoscore.core.image import Image
from neoscore.core.layout_controllers import MarginController, NewLine
from neoscore.core.music_char import MusicChar
from neoscore.core.music_font import MusicFont
from neoscore.core.music_text import MusicText
from neoscore.core.page import Page
from neoscore.core.page_overlays import simple_header_footer
from neoscore.core.paper import Paper
from neoscore.core.path import Path
from neoscore.core.pen import Pen
from neoscore.core.pen_cap_style import PenCapStyle
from neoscore.core.pen_join_style import PenJoinStyle
from neoscore.core.pen_pattern import PenPattern
from neoscore.core.point import ORIGIN, Point
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.raw_music_char import RawMusicChar
from neoscore.core.rect import Rect
from neoscore.core.repeating_music_text_line import RepeatingMusicTextLine
from neoscore.core.rich_text import RichText
from neoscore.core.text import Text
from neoscore.core.text_alignment import AlignmentX, AlignmentY
from neoscore.core.units import ZERO, Inch, Mm, Unit
from neoscore.western import barline_style, notehead_tables
from neoscore.western.accidental import Accidental
from neoscore.western.accidental_type import AccidentalType
from neoscore.western.arpeggio_line import ArpeggioLine
from neoscore.western.barline import Barline
from neoscore.western.barline_style import BarlineStyle
from neoscore.western.beam_group import BeamGroup
from neoscore.western.brace import Brace
from neoscore.western.chordrest import Chordrest
from neoscore.western.clef import Clef
from neoscore.western.clef_type import ClefType
from neoscore.western.duration import Duration
from neoscore.western.dynamic import Dynamic
from neoscore.western.flag import Flag
from neoscore.western.hairpin import Hairpin
from neoscore.western.instrument_name import InstrumentName
from neoscore.western.invisible_clef import InvisibleClef
from neoscore.western.key_signature import KeySignature
from neoscore.western.key_signature_type import KeySignatureType
from neoscore.western.ledger_line import LedgerLine
from neoscore.western.meter import COMMON_TIME, CUT_TIME, Meter
from neoscore.western.metronome_mark import MetronomeMark
from neoscore.western.notehead import Notehead
from neoscore.western.octave_line import OctaveLine
from neoscore.western.ped_and_star import PedAndStar
from neoscore.western.pedal_line import PedalLine
from neoscore.western.pitch import Pitch
from neoscore.western.rest import Rest
from neoscore.western.slur import Slur
from neoscore.western.staff import Staff
from neoscore.western.staff_group import StaffGroup
from neoscore.western.stem import Stem
from neoscore.western.system_line import SystemLine
from neoscore.western.tab_clef import TabClef
from neoscore.western.tab_number import TabNumber
from neoscore.western.tab_staff import TabStaff
from neoscore.western.tab_string_text import TabStringText
from neoscore.western.tie import Tie
from neoscore.western.time_signature import TimeSignature
from neoscore.western.tuplet import Tuplet



import os
import pathlib
from typing import Union


def resolve_qt_path(path: Union[str, pathlib.Path]) -> str:
    """Convert a path to a string compatible with Qt's path-using methods."""
    if isinstance(path, str):
        path = pathlib.Path(path)
    return os.path.abspath(path)



from enum import Enum


class PenCapStyle(Enum):

    """Styles controlling how pen strokes are capped.

    The enum int values are for internal purposes and not guaranteed by the API.
    """

    # Values align with Qt's http://doc.qt.io/qt-5.9/qt.html#PenCapStyle-enum

    SQUARE = 0x10
    """A square cap that extends beyond the end point by half the pen width"""

    FLAT = 0x00
    """A square cap that does not cover the end point of the path"""

    ROUND = 0x20
    """A rounded cap"""



from dataclasses import dataclass, field

from neoscore.core.units import ZERO, Unit


@dataclass(frozen=True)
class StaffFringeLayout:

    """Layout specification for a :obj:`.Staff` fringe.

    Each field except ``pos_x_in_staff`` describes an X-axis position relative to the
    start of the staff's live area. In flowables, that start point is typically the
    beginning of the flowable :obj:`.NewLine`. Outside flowables, that start point is
    the X-axis position of the staff.

    Since fringes extend to the left of this reference point, the values here will
    typically be negative.
    """

    pos_x_in_staff: Unit
    """The staff position of the line where this fringe is placed.

    In staves outside flowables, this will always be ``ZERO``.
    """

    staff: Unit
    """The starting position of the drawn staff in the fringe."""

    clef: Unit
    """The clef's position"""

    key_signature: Unit
    """The key signature's position, if any.

    If there is no key signature this will be ``ZERO``.
    """

    time_signature: Unit = field(default_factory=lambda: ZERO)
    """The time signature's position, if any.

    If there is no time signature this will be ``ZERO``.
    """



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Optional, Union

from typing_extensions import TypeAlias

from neoscore.core.exceptions import InvalidPitchDescriptionError
from neoscore.western.accidental_type import AccidentalType


@dataclass(frozen=True)
class Pitch:

    """A written pitch with a letter, octave, and accidental.

    This class does not define an actual concert pitch, MIDI code, pitch class, etc.
    associated with it. Users building notation systems on it can decide whether this
    represents a concert pitch or a written one. Neoscore's ``western`` module treats it
    mostly as a written pitch, unconditionally writing provided accidentals regardless
    of context and key signatures.

    The class supports a helpful shorthand for standard western 12-EDO pitches inspired
    by Lilypond's pitch notation; see :obj:`.Pitch.from_str`.

    Extended accidentals are fully supported by passing arbitrary SMuFL glyph names to
    the ``accidental`` attribute.
    """

    _shorthand_regex = re.compile("^([a-g])(s|#|n|f|b|ss|x|ff|bb)?('*|,*)$")
    _diatonic_degrees_in_c = {"c": 1, "d": 2, "e": 3, "f": 4, "g": 5, "a": 6, "b": 7}
    _middle_c_octave = 4
    _accidental_shorthands = {
        "s": AccidentalType.SHARP,
        "#": AccidentalType.SHARP,
        "n": AccidentalType.NATURAL,
        "f": AccidentalType.FLAT,
        "b": AccidentalType.FLAT,
        "ss": AccidentalType.DOUBLE_SHARP,
        "x": AccidentalType.DOUBLE_SHARP,
        "ff": AccidentalType.DOUBLE_FLAT,
        "bb": AccidentalType.DOUBLE_FLAT,
    }

    letter: str
    """The a-g letter name of the pitch."""

    accidental: Optional[AccidentalType | str]
    """An accidental associated with the pitch.

    For conventional accidentals, this can be an ``AccidentalType``.
    Alternatively, this can be an arbitrary SMuFL glyph name string.
    """

    octave: int
    """The octave number, where 4 is the octave starting with middle-C."""

    @classmethod
    def from_str(cls, shorthand: str) -> Pitch:
        """Create a conventional Western ``Pitch`` from a string shorthand.

        The pitch shorthand is inspired by Lilypond's. It consists of three parts:
        a pitch letter, an optional accidental, and an optional octave mark.

        * Pitch letters are the standard ``c`` through ``b`` letters.
        * The accidental may be ``f`` or ``b`` for flat, ``s`` or ``#`` for sharp, ``n`` for
          natural, ``ss`` or ``x`` for double-sharp, and ``ff`` or ``bb`` for double-flat.
        * The octave indication is given by a series of apostrophes (``'``)
          or commas (``,``), where each apostrophe increases the pitch by an octave,
          and each comma decreases it. All octave transformations are relative to
          the octave starting at middle-C. The absence of an octave indicator means a
          pitch is within the octave starting at middle-C. (Note that this differs from
          Lilypond's notation, which starts at the octave *below* middle-C.)

        Some examples:

        * Middle-C: ``c``
        * The B directly below that: ``b,``
        * The C one octave below middle-C: ``c,``
        * The E-flat above middle-C: ``ef`` or ``eb``
        * The F-sharp above middle-C: ``fs`` or ``f#``
        * The G-double-sharp above middle-C: ``fx`` or ``fss``
        * The A-double-flat above the treble staff: ``aff'`` or ``abb'``

        """
        match = Pitch._shorthand_regex.match(shorthand)
        if match is None:
            raise InvalidPitchDescriptionError
        letter = match.group(1)
        accidental_str = match.group(2)
        ticks = match.group(3)
        letter = letter
        if accidental_str:
            accidental = Pitch._accidental_shorthands[accidental_str.lower()]
        else:
            accidental = None
        octave = 4
        if ticks:
            octave += len(ticks) * (-1 if ticks[0] == "," else 1)
        return Pitch(letter, accidental, octave)

    @classmethod
    def from_def(cls, pitch_def: PitchDef) -> Pitch:
        if isinstance(pitch_def, Pitch):
            return pitch_def
        elif isinstance(pitch_def, tuple):
            return Pitch(*pitch_def)
        return Pitch.from_str(pitch_def)

    @property
    def diatonic_degree_in_c(self) -> int:
        """The diatonic degree of the pitch as if it were in C.

        >>> Pitch.from_str("c").diatonic_degree_in_c
        1
        >>> Pitch.from_str("c'").diatonic_degree_in_c
        1
        >>> Pitch.from_str("d'''").diatonic_degree_in_c
        2
        >>> Pitch.from_str("bf,").diatonic_degree_in_c
        7
        """
        return Pitch._diatonic_degrees_in_c[self.letter]

    @property
    def staff_pos_from_middle_c(self) -> float:
        """The pitch's staff position relative to middle C.

        Values are in numeric pseudo-staff-units where positive
        values mean positions below middle C, and negative values
        mean positions above it.

        >>> Pitch.from_str("c").staff_pos_from_middle_c
        0
        >>> Pitch.from_str("cs").staff_pos_from_middle_c
        0
        >>> Pitch.from_str("d").staff_pos_from_middle_c
        -0.5
        >>> Pitch.from_str("d'").staff_pos_from_middle_c
        -4
        >>> Pitch.from_str("cn,,").staff_pos_from_middle_c
        7
        """
        middle_c = (4 * 7) + 1  # C at octave 4
        note_pos = (self.octave * 7) + self.diatonic_degree_in_c
        position = (note_pos - middle_c) / -2
        if position % 1 == 0:
            return int(position)
        else:
            return position


PitchDef: TypeAlias = Union[Pitch, str, tuple]
"""Shorthand for a ``Pitch``

May be either a ``Pitch``, a pitch string shorthand (see ``Pitch.from_str``), or a ``Pitch``
init arg tuple.
"""



"""Various interoperable units classes and some related helper functions."""
from __future__ import annotations

import decimal
from typing import Any, Optional, Type, TypeVar, Union, cast

TUnit = TypeVar("TUnit", bound="Unit")


class Unit:
    """An immutable graphical distance with a unit.

    Unit objects enable easy conversion from one unit to another and convenient
    operations between them.

    Common operators (``+``, ``-``, ``/``, etc.) are supported between them. Return
    values from these operations are given in the type on the left.

        >>> from neoscore.core.units import Inch, Mm
        >>> print(Inch(1) + Mm(1))
        Inch(1.039)

    To facilitate easy comparison between equivalent values in different units, equality
    is checked with a tolerance of ``Unit(0.001)``.

        >>> from neoscore.core.units import Inch, Mm
        >>> assert Inch(Mm(1)) == Mm(1)
        >>> assert Inch(Mm(1)) >= Mm(1)
        >>> assert Inch(Mm(1)) <= Mm(1)

    The base ``Unit`` type is a graphical unit corresponding to 1/72nd of an inch.
    Internally, this is significant because it is the unit used by the graphics backend,
    Qt. For most purposes, you probably want to use a more descriptive unit type.
    """

    __slots__ = {
        "base_value": "The underlying float value in base units.",
        "_display_value": "",
    }

    CONVERSION_RATE: float = 1
    """The ratio of this class to fundamental ``Unit``\ s.

    Subclasses should override this.
    """

    def __init__(self, value: Unit | float, _raw_base_value=None):
        """Create a unit from another unit or a raw number."""
        if _raw_base_value is not None:
            # Short-circuiting constructor for internal use
            self.base_value = _raw_base_value
            self._display_value = None
        else:
            base_value = getattr(value, "base_value", None)
            if base_value is not None:
                self.base_value = base_value
                self._display_value = None
            else:
                self.base_value = value * self.CONVERSION_RATE
                self._display_value = value

    @property
    def display_value(self) -> float:
        """A human-friendly unit value.

        If the unit was constructed with a simple number, this will return the exact
        given argument value. If the unit was constructed from another unit, this will
        return the converted value rounded to 3 decimal places.
        """
        if self._display_value:
            return self._display_value
        return round(self.base_value / self.CONVERSION_RATE, 3)

    @property
    def rounded_base_value(self) -> float:
        """The base value rounded to 2 decimal places.

        This is useful for things like hash keys for caching purposes
        """
        return round(self.base_value, 2)

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.display_value)

    _CMP_POS_EPSILON = 0.001
    _CMP_NEG_EPSILON = -0.001

    def __lt__(self, other: Unit):
        return self.base_value - other.base_value < Unit._CMP_NEG_EPSILON

    def __le__(self, other: Unit):
        return self.base_value - other.base_value < Unit._CMP_POS_EPSILON

    def __eq__(self, other: Any):
        """Two units are equal if the difference between their base
        values is less than 0.001, or if both are infinite.
        """
        return hasattr(other, "base_value") and (
            # Check strict equality first for speed and coverage of infinity values
            self.base_value == other.base_value
            or abs(self.base_value - other.base_value) < Unit._CMP_POS_EPSILON
        )

    def __gt__(self, other: Unit):
        return self.base_value - other.base_value > Unit._CMP_POS_EPSILON

    def __ge__(self, other: Unit):
        return self.base_value - other.base_value > Unit._CMP_NEG_EPSILON

    def __add__(self: TUnit, other: TUnit) -> TUnit:
        return type(self)(None, _raw_base_value=self.base_value + other.base_value)

    def __sub__(self: TUnit, other: Unit) -> TUnit:
        return type(self)(None, _raw_base_value=self.base_value - other.base_value)

    def __mul__(self: TUnit, other: float) -> TUnit:
        if hasattr(other, "base_value"):
            raise TypeError
        return type(self)(None, _raw_base_value=self.base_value * other)

    def __rmul__(self: TUnit, other: float) -> TUnit:
        # __rmul__ behaves identically to __mul__
        if hasattr(other, "base_value"):
            raise TypeError
        return type(self)(None, _raw_base_value=self.base_value * other)

    def __truediv__(self: TUnit, other: Union[Unit, float]) -> Union[TUnit, float]:
        if hasattr(other, "base_value"):
            # Unit / Unit -> Float
            return self.base_value / (cast(Unit, other)).base_value
        # Unit / Float -> Unit
        return type(self)(None, _raw_base_value=self.base_value / other)

    def __pow__(self: TUnit, other: float, modulo: Optional[int] = None) -> TUnit:
        return type(self)(Unit(pow(self.base_value, other, modulo)))

    def __neg__(self: TUnit) -> TUnit:
        return type(self)(None, _raw_base_value=-self.base_value)

    def __abs__(self: TUnit) -> TUnit:
        return type(self)(None, _raw_base_value=abs(self.base_value))


class Inch(Unit):
    """An inch."""

    CONVERSION_RATE = 72


class Mm(Unit):
    """A millimeter."""

    CONVERSION_RATE = float(decimal.Decimal("0.0393700787") * Inch.CONVERSION_RATE)


ZERO = Unit(0)
"""Shorthand for a zero unit"""


def make_unit_class(name: str, conversion_rate: float) -> Type[Unit]:
    """Create a ``Unit`` subclass with a name and base unit conversion rate"""
    return cast(
        Type[Unit],
        type(
            name,
            (Unit,),
            {"CONVERSION_RATE": conversion_rate},
        ),
    )


def _convert_all_to_unit_out_of_place(
    collection: Union[tuple, set], unit: Type[Unit]
) -> Union[tuple, set]:
    mutable_iterable: List[Any] = list(collection)
    convert_all_to_unit(mutable_iterable, unit)
    return type(collection)(mutable_iterable)


def convert_all_to_unit(collection: Union[list, dict], unit: Type[Unit]):
    """Recursively convert all numbers found in a list or dict to a unit in place.

    This function works in place. Tuples and sets found within ``collection`` will be
    replaced.

    In dictionaries, only values are converted.
    """
    if isinstance(collection, dict):
        for key, value in collection.items():
            if isinstance(value, (int, float, Unit)):
                collection[key] = unit(value)
            elif isinstance(value, (list, dict)):
                convert_all_to_unit(collection[key], unit)
            elif isinstance(value, (tuple, set)):
                collection[key] = _convert_all_to_unit_out_of_place(
                    collection[key], unit
                )
            # (else: continue --- nothing to do here)
    elif isinstance(collection, list):
        for i in range(len(collection)):
            if isinstance(collection[i], (int, float, Unit)):
                collection[i] = unit(collection[i])
            elif isinstance(collection[i], (list, dict)):
                convert_all_to_unit(collection[i], unit)
            elif isinstance(collection[i], (tuple, set)):
                collection[i] = _convert_all_to_unit_out_of_place(collection[i], unit)
            # (else: continue --- nothing to do here)
    else:
        raise TypeError



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from sortedcontainers import SortedKeyList  # type: ignore

from neoscore.core import neoscore
from neoscore.core.layout_controllers import NewLine
from neoscore.core.units import ZERO, Unit
from neoscore.western.staff_fringe_layout import StaffFringeLayout

if TYPE_CHECKING:
    from neoscore.western.abstract_staff import AbstractStaff


class StaffGroup:

    """A collection of staves.

    This provides a shared context for staff systems, useful in situations like aligning
    staff fringes.

    Each staff has a reference to a staff group, even if it's a single-item group. See
    :obj:`.AbstractStaff.group`.
    """

    # Fringe padding constants in pseudo-staff-units
    RIGHT_PADDING = 1
    """Padding to the right of all staff fringes, in pseudo-staff-units."""

    CLEF_LEFT_PADDING = 0.5
    """Padding to the left of clefs in fringes, in pseudo-staff-units."""

    TIME_SIG_LEFT_PADDING = 0.5
    """Padding to the left of time signatures in fringes, in pseudo-staff-units."""

    KEY_SIG_LEFT_PADDING = 0.25
    """Padding to the left of key signatures in fringes, in pseudo-staff-units."""

    def __init__(self) -> None:
        self._fringe_layout_cache: Dict[
            Tuple[AbstractStaff, Optional[NewLine]], StaffFringeLayout
        ] = {}
        # Sort staves according to position to some arbitrary known object
        self._staves: SortedKeyList[AbstractStaff] = SortedKeyList(
            key=lambda s: neoscore.document.pages[0].map_to(s).y
        )

    @property
    def staves(self) -> SortedKeyList[AbstractStaff]:
        """The staves contained in this group.

        Staves shouldn't be directly added to this list; instead register a staff with
        the group when the staff is initialized.

        This list is automatically sorted in visually descending order.
        """
        return self._staves

    def fringe_layout_at(
        self, staff: AbstractStaff, location: Optional[NewLine]
    ) -> StaffFringeLayout:
        """Compute to fringe layout needed for a staff at a given location.

        This automatically aligns the returned layout with the fringes of other staves
        in the group.

        Layouts are internally cached so this is fairly efficient.
        """
        # If the layout is already cached, return it
        cached_value = self._fringe_layout_cache.get((staff, location))
        if cached_value:
            return cached_value
        # If the staff doesn't actually exist at the beginning of a line,
        # Do its fringe layout in isolation
        if not self._staff_exists_at(staff, location):
            layout = staff.fringe_layout_for_isolated_staff(location)
            self._fringe_layout_cache[(staff, location)] = layout
            return layout
        # Work out the layouts of each staff in isolation first
        isolated_layouts = []
        min_staff_basis = ZERO  # Wider fringes give a lower number here (further left)
        for iter_staff in self.staves:
            if not self._staff_exists_at(iter_staff, location):
                continue
            layout = iter_staff.fringe_layout_for_isolated_staff(location)
            min_staff_basis = min(min_staff_basis, layout.staff)
            isolated_layouts.append((iter_staff, layout))
        # Then align each layout to fit the widest layout found, and store each in cache
        for iter_staff, isolated_layout in isolated_layouts:
            aligned_layout = self._align_layout(isolated_layout, min_staff_basis)
            self._fringe_layout_cache[(iter_staff, location)] = aligned_layout
        # Now the requested layout is cached, so return it
        return self._fringe_layout_cache[(staff, location)]

    @staticmethod
    def _staff_exists_at(staff: AbstractStaff, location: Optional[NewLine]) -> bool:
        """Check if a staff exists at a given location"""
        if not location:
            # No good way currently to check without a location
            return True
        staff_pos_x = staff.flowable.descendant_pos_x(staff)
        return (staff_pos_x <= location.flowable_x) and (
            staff_pos_x + staff.breakable_length >= location.flowable_x
        )

    @staticmethod
    def _align_layout(
        layout: StaffFringeLayout, staff_basis: Unit
    ) -> StaffFringeLayout:
        if layout.staff == staff_basis:
            return layout
        # Find the delta from the layout's staff basis to the alignment's
        # Note that these are all *negative* numbers
        delta = staff_basis - layout.staff
        return StaffFringeLayout(
            layout.pos_x_in_staff,
            # Staff, clef, and key signatures are left-aligned to the widest fringe
            layout.staff + delta,
            layout.clef + delta,
            layout.key_signature + delta,
            # Time signatures are right-aligned
            layout.time_signature,
        )



from typing import Iterable, List, Optional

from neoscore.core.break_hint import BreakHint
from neoscore.core.color import ColorDef
from neoscore.core.has_music_font import HasMusicFont
from neoscore.core.music_font import MusicFont
from neoscore.core.path import Path
from neoscore.core.pen import Pen
from neoscore.core.pen_pattern import PenPattern
from neoscore.core.point import ORIGIN, Point
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import ZERO, Union, Unit
from neoscore.western import barline_style
from neoscore.western.abstract_staff import AbstractStaff
from neoscore.western.barline_style import BarlineStyle
from neoscore.western.multi_staff_object import MultiStaffObject
from neoscore.western.staff_group import StaffGroup

_DEFAULT_BARLINE_STYLE = BarlineStyle()


class Barline(PositionedObject, MultiStaffObject, HasMusicFont):
    """A barline spanning any number of staves.

    In addition to a default plain barline, this supports many common barline styles out
    of the box, and beyond those you can easily define custom barline styles.

    A :obj:`.BreakHint` is automatically attached to the end of the barline so
    :obj:`.Flowable` containers will treat barlines as potential line break points.
    """

    def __init__(
        self,
        pos_x: Unit,
        staves: Union[StaffGroup, List[AbstractStaff]],
        styles: Union[BarlineStyle, Iterable[BarlineStyle]] = barline_style.SINGLE,
        connected: Optional[bool] = True,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            pos_x: The barline X position relative to the highest staff.
                Specifies right edge of the barline group and offsets 'thickness'.
            staves: The staves spanned. If a raw list of staves is given, it must be
                in descending order.
            styles: This accepts any of the pre-made styles provided in
                :obj:`.barline_style`, in addition to custom styles defined in
                a list of :obj:`.BarlineStyle`\ s (one for each sub-barline).
            connected: Whether to connect the barline between staves.
            font: If provided, this overrides the font in the parent (top) staff.
        """
        MultiStaffObject.__init__(self, staves)
        PositionedObject.__init__(self, (pos_x, ZERO), self.highest)

        if font is None:
            font = HasMusicFont.find_music_font(self.highest)
        self._music_font = font
        self.engraving_defaults = self._music_font.engraving_defaults
        self.paths = []
        self.connected = connected

        # Start x position for first barline relative to self
        start_x = ZERO

        if isinstance(styles, BarlineStyle):
            styles = [styles]

        fallback_thickness = self.unit(_DEFAULT_BARLINE_STYLE.thickness)
        fallback_gap_right = self.unit(_DEFAULT_BARLINE_STYLE.gap_right)

        # draw each of the bar lines in turn from right to left
        for i in reversed(range(len(styles))):
            style = styles[i]
            thickness = self._resolve_style_measurement(
                style.thickness, fallback_thickness
            )
            # adjust start x to accommodate pen thickness
            start_x -= thickness / 2
            self._draw_barline(start_x, thickness, style.pattern, style.color)
            # move to next line to the left
            if i != 0:
                start_x -= thickness + self._resolve_style_measurement(
                    styles[i - 1].gap_right, fallback_gap_right
                )

        # Attach a break hint at the edge of the rightmost barline
        self._break_hint = BreakHint(ORIGIN, self)

    @property
    def music_font(self) -> MusicFont:
        return self._music_font

    def _draw_barline(
        self, start_x: Unit, thickness: Unit, pen_pattern: PenPattern, color: ColorDef
    ):
        # Create the path
        line_path = Path(
            Point(start_x, ZERO),
            self,
            pen=Pen(pattern=pen_pattern, thickness=thickness, color=color),
        )

        # Draw the path
        if self.connected:
            line_path.move_to(ZERO, self.highest.barline_extent[0])
            line_path.line_to(
                ZERO, self.map_to(self.lowest).y + self.lowest.barline_extent[1]
            )

        else:
            y_offset = self.highest.y
            for stave in self.staves:
                new_y = stave.pos.y - y_offset
                line_path.move_to(ZERO, new_y + stave.barline_extent[0])
                line_path.line_to(ZERO, new_y + stave.barline_extent[1])

        self.paths.append(line_path)

    def _resolve_style_measurement(
        self, value: Union[str, float, Unit], fallback: Unit
    ) -> Unit:
        if isinstance(value, Unit):
            return value
        elif isinstance(value, (int, float)):
            return self.unit(value)
        else:
            return self.engraving_defaults.get(value, fallback)



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Dict, Optional, Tuple, Union

from typing_extensions import TypeAlias

KeySignatureLayout: TypeAlias = Dict[str, Tuple[float, float]]
"""A layout specification for accidentals in key signatures.

Instances should have a key for every pitch letter a-g (lowercase),
and each value should be an ``(x, y)`` tuple of pseudo staff positions
to be plugged into a staff's unit.
"""


StaffPosFunc: TypeAlias = Callable[[int], float]
"""A function which takes a number of staff lines and returns a staff position."""


@dataclass(frozen=True)
class ClefType:
    """A logical clef specifier.

    Many standard clefs are pre-defined in this module, but users can
    also create custom clef types using this class.
    """

    glyph_name: str
    """The SMuFL glyph name used to represent the clef."""

    """
    how to deal with percussion clefs properly?

    their staff position and middle C position both need to be
    dynamically calculated from staff line count

    """

    staff_pos: Union[float, StaffPosFunc]
    """Where the clef should be vertically drawn in a staff.

    This is given in pseudo staff units relative to the staff's top
    line. When positioning clefs, this should be passed into the
    staff's unit to find the appropriate y position.

    For clefs whose position depends on the number of staff lines,
    this can be a ``StaffPosFunc``.
    """

    middle_c_staff_pos: Union[float, StaffPosFunc]
    """Where this clef places middle C in a staff.

    Like ``staff_pos``, this is given in pseudo staff units relative to
    the staff's top line.

    For clefs whose middle C position depends on the number of staff
    lines, this can be a ``StaffPosFunc``.
    """

    key_signature_flat_layout: Optional[KeySignatureLayout]
    """Set of positions for flat key sigantures, if applicable."""

    key_signature_sharp_layout: Optional[KeySignatureLayout]
    """Set of positions for sharp key sigantures, if applicable."""

    @classmethod
    def from_def(cls, clef_type_def: ClefTypeDef) -> ClefType:
        if isinstance(clef_type_def, ClefType):
            return clef_type_def
        return CLEF_TYPE_SHORTHAND_NAMES[clef_type_def.lower()]


_sharp_positions: KeySignatureLayout = {
    "f": (0, 0),
    "c": (1, 1.5),
    "g": (2, -0.5),
    "d": (3, 1),
    "a": (4, 2.5),
    "e": (5, 0.5),
    "b": (6, 2),
}
_flat_positions: KeySignatureLayout = {
    "b": (0, 2),
    "e": (1, 0.5),
    "a": (2, 2.5),
    "d": (3, 1),
    "g": (4, 3),
    "c": (5, 1.5),
    "f": (6, 3.5),
}
_bass_flat_positions = {
    key: (value[0], value[1] + 1) for key, value in _flat_positions.items()
}
_tenor_flat_positions = {
    key: (value[0], value[1] - 0.5) for key, value in _flat_positions.items()
}
_alto_flat_positions = {
    key: (value[0], value[1] + 0.5) for key, value in _flat_positions.items()
}
_bass_sharp_positions = {
    key: (value[0], value[1] + 1) for key, value in _sharp_positions.items()
}
_tenor_sharp_positions = {
    "f": (0, 3),
    "c": (1, 1),
    "g": (2, 2.5),
    "d": (3, 0.5),
    "a": (4, 2),
    "e": (5, 0),
    "b": (6, 1.5),
}
_alto_sharp_positions = {
    key: (value[0], value[1] + 0.5) for key, value in _sharp_positions.items()
}


TREBLE = ClefType("gClef", 3, 5, _flat_positions, _sharp_positions)
"""A conventional treble clef.

:meta hide-value:
"""

TREBLE_8VB = ClefType("gClef8vb", 3, 1.5, _flat_positions, _sharp_positions)
"""A treble clef one octave below, like often used in tenor vocal parts.

:meta hide-value:
"""

TREBLE_8VA = ClefType("gClef8va", 3, 8.5, _flat_positions, _sharp_positions)
"""A treble clef one octave below, like often used in tenor vocal parts.

:meta hide-value:
"""

TREBLE_8VB = ClefType("gClef8vb", 3, 1.5, _flat_positions, _sharp_positions)
"""A treble clef one octave below, like often used in tenor vocal parts.

:meta hide-value:
"""

BASS = ClefType("fClef", 1, -1, _bass_flat_positions, _bass_sharp_positions)
"""A conventional bass clef.

:meta hide-value:
"""

BASS_8VB = ClefType("fClef8vb", 1, -4.5, _bass_flat_positions, _bass_sharp_positions)
"""A bass clef one octave below, like often used in double bass parts.

:meta hide-value:
"""

TENOR = ClefType("cClef", 1, 1, _tenor_flat_positions, _tenor_sharp_positions)
"""A tenor C-clef.

:meta hide-value:
"""

ALTO = ClefType("cClef", 2, 2, _alto_flat_positions, _alto_sharp_positions)
"""An alto C-clef.

:meta hide-value:
"""


def _percussion_staff_pos_func(num_lines: int) -> float:
    return (num_lines - 1) / 2


PERCUSSION_1 = ClefType(
    "unpitchedPercussionClef1",
    _percussion_staff_pos_func,
    _percussion_staff_pos_func,
    _alto_flat_positions,
    _alto_sharp_positions,
)
"""A percussion clef consisting of 2 solid bars.

Percussion clefs are treated as C clefs always centered at the middle
of the staff. This works with any number of staff lines.

:meta hide-value:
"""

PERCUSSION_2 = ClefType(
    "unpitchedPercussionClef2",
    _percussion_staff_pos_func,
    _percussion_staff_pos_func,
    _alto_flat_positions,
    _alto_sharp_positions,
)
"""A percussion clef consisting of an open rectangle.

See also ``PERCUSSION_1``

:meta hide-value:
"""

BRIDGE = ClefType(
    "bridgeClef",
    _percussion_staff_pos_func,
    _percussion_staff_pos_func,
    _alto_flat_positions,
    _alto_sharp_positions,
)
"""A bridge clef.

Like percussion clefs, the bridge clef is treated as a C clef always centered at the
middle of the staff. This works with any number of staff lines.

:meta hide-value:
"""

CLEF_TYPE_SHORTHAND_NAMES = {
    "treble": TREBLE,
    "treble_8vb": TREBLE_8VB,
    "treble_8va": TREBLE_8VA,
    "bass": BASS,
    "bass_8vb": BASS_8VB,
    "tenor": TENOR,
    "alto": ALTO,
    "percussion_1": PERCUSSION_1,
    "percussion_2": PERCUSSION_2,
    "bridge": BRIDGE,
}

ClefTypeDef: TypeAlias = Union[ClefType, str]
"""ClefTypes can be given by the string name of any pre-defined ClefType in this module.

String lookup is case-insensitive.
"""



from __future__ import annotations

from typing import Optional

from sortedcontainers import SortedKeyList

from neoscore.core import neoscore
from neoscore.core.layout_controllers import MarginController, NewLine
from neoscore.core.point import Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.units import ZERO, Mm, Unit


class Flowable(PositionedObject):

    """A flowable coordinate space container.

    This provides a virtual horizontal strip of space in which objects can be placed,
    and at render time be automatically flowed across line breaks and page breaks in the
    document.

    To place an object in a ``Flowable``, simply parent it to one, or to an object
    already in one.

    In typical scores, there will be a single ``Flowable`` placed in the first page of
    the document, and most objects will be placed inside it.
    """

    _neoscore_flowable_type_marker = True

    def __init__(
        self,
        pos: PointDef,
        parent: Optional[PositionedObject],
        length: Unit,
        height: Unit,
        y_padding: Unit = Mm(5),
        break_threshold: Unit = Mm(5),
    ):
        """
        Args:
            pos: Starting position in relative to the top left corner of the
                live document area of the first page
            parent: An optional parent object. Nested flowables are not supported,
                so this should not be a flowable or in one. This defaults to the
                document's first page.
            length: length of the flowable
            height: height of the flowable
            y_padding: The vertical gap between flowable sections
            break_threshold: The maximum distance the flowable will shorten a line
                to allow a break to occur on a ``BreakOpportunity``
        """
        super().__init__(pos, parent)
        self._length = length
        self._height = height
        self._y_padding = y_padding
        self._break_threshold = break_threshold
        self._lines = []
        self._provided_controllers = Flowable._new_provided_controllers_list()

    @property
    def length(self) -> Unit:
        """The length of the unwrapped flowable"""
        return self._length

    @property
    def height(self) -> Unit:
        """The height of the unwrapped flowable"""
        return self._height

    @height.setter
    def height(self, value: Unit):
        self._height = value

    @property
    def y_padding(self) -> Unit:
        """The padding between wrapped sections of the flowable"""
        return self._y_padding

    @y_padding.setter
    def y_padding(self, value: Unit):
        self._y_padding = value

    @property
    def break_threshold(self) -> Unit:
        """The threshold for :obj:`.BreakOpportunity`-aware line breaks.

        This is the maximum distance the flowable will shorten a line to allow
        a break to occur on a ``BreakOpportunity``.

        If set to ``ZERO``, break opportunities will be entirely ignored during
        layout. On the other hand, if set to a value larger than the live page width,
        all break opportunities will be taken.
        """
        return self._break_threshold

    @break_threshold.setter
    def break_threshold(self, value: Unit):
        self._break_threshold = value

    @property
    def lines(self) -> List[NewLine]:
        """The generated lines of this flowable.

        This property is managed and should not be modified."""
        return self._lines

    @lines.setter
    def lines(self, value: List[NewLine]):
        self._lines = value

    @property
    def provided_controllers(self) -> SortedKeyList[MarginController]:
        """Layout controllers provided by users.

        Currently, this only supports margin controllers. Eventually on we may expand
        this to allow things like explicit user-defined ``NewLine``\ s.

        Controllers should not be added directly to this list; use
        :obj:`.add_margin_controller` instead.
        """
        return self._provided_controllers

    def add_margin_controller(self, controller: MarginController):
        """Add a margin controller if applicable.

        If ``provided_controllers`` already has a margin controller at the given
        ``controller.flowable_x`` with the same ``layer_key``, the controller is only
        inserted if its margin is larger than the existing one.
        """
        if not self._provided_controllers:
            self._provided_controllers.add(controller)
            return
        idx = self._provided_controllers.bisect_left(controller)

        for i in range(idx, len(self._provided_controllers)):
            existing_controller = self._provided_controllers[i]
            if existing_controller.flowable_x > controller.flowable_x:
                # No existing controllers at this flowable_x and layer found
                break
            if existing_controller.layer_key == controller.layer_key:
                # Existing controller at this flowable_x and layer found
                if existing_controller.margin_left >= controller.margin_left:
                    # Existing controller margin is larger than one being added; skip it
                    return
                else:
                    # Existing controller margin is smaller than one being added; replace it
                    self._provided_controllers.pop(i)
                    break
        self._provided_controllers.add(controller)

    def _generate_lines(self):
        """Generate automatic layout controllers.

        The generated controllers are stored in ``self.layout_controllers``
        in sorted order by ascending x position
        """
        live_page_width = neoscore.document.paper.live_width
        live_page_height = neoscore.document.paper.live_height
        break_opps = self._find_break_opportunities()
        for c in self.lines:
            c.remove()
        self.lines = []
        while True:
            if not self.lines:
                flowable_start_x = ZERO
                page = self.first_ancestor_with_attr("_neoscore_page_type_marker")
                flowable_page_pos = page.map_to(self)
                new_line_x = flowable_page_pos.x + self._active_margin_at(
                    flowable_start_x
                )
                new_line_y = flowable_page_pos.y
            else:
                last = self.lines[-1]
                flowable_start_x = last.flowable_x + last.length
                page = last.page
                new_line_x = self._active_margin_at(flowable_start_x)
                new_line_y = last.y + self.height + self.y_padding
                new_line_bottom_y = new_line_y + self.height
                if (
                    new_line_y > live_page_height
                    or new_line_bottom_y > live_page_height
                ):
                    page = neoscore.document.pages[page.index + 1]
                    new_line_y = ZERO
            # Now determine this line's length
            max_length = live_page_width - new_line_x
            max_line_end_flowable_x = flowable_start_x + max_length
            nearest_break_opp = next(
                (opp for opp in reversed(break_opps) if opp < max_line_end_flowable_x),
                None,
            )
            if (
                nearest_break_opp
                and max_line_end_flowable_x - nearest_break_opp < self.break_threshold
            ):
                length = nearest_break_opp - flowable_start_x
            else:
                length = max_length
            self.lines.append(
                NewLine(
                    (new_line_x, new_line_y),
                    page,
                    flowable_start_x,
                    length,
                    self.height,
                )
            )
            if flowable_start_x + length > self.length:
                break

    def _find_break_opportunities(self) -> List[Unit]:
        """Find the relative X positions of every break hint in this flowable.

        The returned positions will be sorted.
        """
        opps = self.descendants_with_attribute(
            "_neoscore_break_opportunity_type_marker"
        )
        return sorted((self.map_x_to(opp) for opp in opps))

    def _active_margin_at(self, flowable_x: Unit) -> Unit:
        active_margin_layers: Dict[str, Unit] = {}
        for controller in self.provided_controllers:
            if controller.flowable_x > flowable_x:
                break
            active_margin_layers[controller.layer_key] = controller.margin_left
        return sum(active_margin_layers.values(), ZERO)

    def map_to_canvas(self, local_point: Point) -> Point:
        """Convert a local point to its position in the canvas.

        Note that this should only be called at render-time, since it depends on layout
        controllers only generated once the flowable has started rendering.

        Args:
            local_point: A position in the flowable's local space.
        """
        if not getattr(self, "_currently_rendering", None):
            print("WARNING: Called Flowable.map_to_canvas outside rendering context")
        line = self.last_break_at(local_point.x)
        line_canvas_pos = line.canvas_pos()
        return line_canvas_pos + Point(local_point.x - line.flowable_x, local_point.y)

    def last_break_at(self, flowable_x: Unit) -> NewLine:
        """Find the last ``NewLine`` that occurred before a given local flowable_x-pos

        Args:
            flowable_x: An x-axis location in the virtual flowable space.
        """
        return self.lines[self.last_break_index_at(flowable_x)]

    def last_break_index_at(self, flowable_x: Unit) -> int:
        """Like ``last_break_at``, but returns an index.

        Args:
            flowable_x: An x-axis location in the virtual flowable space.
        """
        # Note that this assumes that all layout controllers are line
        # breaks, and will not work if/when other types are added
        remaining_x = flowable_x
        for i, controller in enumerate(self.lines):
            remaining_x -= controller.length
            if remaining_x <= ZERO:
                return i
        else:
            return len(self.lines) - 1

    def render(self):
        super().render()

    def pre_render_hook(self):
        super().pre_render_hook()
        self._generate_lines()

    def post_render_hook(self):
        # Clear all auto-generated margin controllers
        super().post_render_hook()
        self._provided_controllers = Flowable._new_provided_controllers_list()

    @staticmethod
    def _new_provided_controllers_list() -> SortedKeyList[MarginController]:
        return SortedKeyList(key=lambda c: c.flowable_x)



from typing import List, Optional, Union, cast

from neoscore.core.brush import BrushDef
from neoscore.core.layout_controllers import MarginController, NewLine
from neoscore.core.music_font import MusicFont, MusicFontGlyphNotFoundError
from neoscore.core.music_text import MusicText
from neoscore.core.pen import PenDef
from neoscore.core.point import Point
from neoscore.core.text_alignment import AlignmentX
from neoscore.core.units import ZERO, Unit
from neoscore.western.abstract_staff import AbstractStaff
from neoscore.western.multi_staff_object import MultiStaffObject
from neoscore.western.staff_group import StaffGroup


class Brace(MultiStaffObject, MusicText):

    """A brace spanning staves.

    This is drawn in the fringe of every staff system in the specified group.
    """

    def __init__(
        self,
        staves: Union[StaffGroup, List[AbstractStaff]],
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """
        Args:
            staves: The staves spanned. If a raw list of staves is given, it must be
                in descending order.
            font: If provided, this overrides the font in the parent (top) staff.
            brush: The brush to fill shapes with.
            pen: The pen to draw outlines with.
        """
        MultiStaffObject.__init__(self, staves)
        # Calculate the height of the brace in highest_staff staff units
        scale = cast(float, self.vertical_span / self.highest.unit(4))
        if self.vertical_span > self.highest.unit(50):
            text = ("brace", 4)
        elif self.vertical_span > self.highest.unit(30):
            text = ("brace", 3)
        elif self.vertical_span > self.highest.unit(15):
            text = ("brace", 2)
        elif self.vertical_span > self.highest.unit(4):
            text = "brace"
        else:
            text = ("brace", 1)
        try:
            # Attempt to use size-specific optional glyph
            MusicText.__init__(
                self,
                (ZERO, self.vertical_span),
                self.highest,
                text,
                font,
                brush,
                pen,
                scale,
                alignment_x=AlignmentX.RIGHT,
            )
        except MusicFontGlyphNotFoundError:
            # Default to non-optional glyph
            MusicText.__init__(
                self,
                (ZERO, self.vertical_span),
                self.highest,
                "brace",
                font,
                brush,
                pen,
                scale,
                alignment_x=AlignmentX.RIGHT,
            )

    @property
    def breakable_length(self) -> Unit:
        """This class's breakable length is that of its highest staff"""
        return self.highest.breakable_length

    def _render_occurrence(
        self,
        pos: Point,
        flowable_line: Optional[NewLine],
    ):
        fringe_layout = self.highest.fringe_layout_at(flowable_line)
        super().render_complete(
            Point(pos.x + fringe_layout.staff, pos.y), flowable_line
        )

    def render_complete(
        self,
        pos: Point,
        flowable_line: Optional[NewLine] = None,
        flowable_x: Optional[Unit] = None,
    ):
        self._render_occurrence(pos, flowable_line)

    def render_before_break(self, pos: Point, flowable_line: NewLine, flowable_x: Unit):
        self._render_occurrence(pos, flowable_line)

    def render_spanning_continuation(
        self, pos: Point, flowable_line: NewLine, object_x: Unit
    ):
        self._render_occurrence(pos, flowable_line)

    def render_after_break(self, pos: Point, flowable_line: NewLine, object_x: Unit):
        self._render_occurrence(pos, flowable_line)

    def _register_layout_controllers(self):
        flowable = self.flowable
        if not flowable:
            return
        staff_flowable_x = flowable.descendant_pos_x(self.highest)
        flowable.add_margin_controller(
            MarginController(
                staff_flowable_x,
                self.bounding_rect.width,
                "_neoscore_brace",
            )
        )

    def pre_render_hook(self):
        super().pre_render_hook()
        self._register_layout_controllers()



from __future__ import annotations

import math
from typing import Optional

from neoscore.core.brush import Brush
from neoscore.core.directions import DirectionX
from neoscore.core.music_font import MusicFont
from neoscore.core.music_path import MusicPath
from neoscore.core.pen import Pen
from neoscore.core.point import Point, PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.spanner_2d import Spanner2D
from neoscore.core.units import ZERO, Unit


class Hairpin(Spanner2D, MusicPath):

    """A crescendo/diminuendo hairpin spanner.

    While this is a path, it requires a music font from which to
    derive its appearance.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        end_pos: PointDef,
        end_parent: Optional[PositionedObject] = None,
        direction: DirectionX = DirectionX.RIGHT,
        width: Optional[Unit] = None,
        font: Optional[MusicFont] = None,
    ):
        """
        Args:
            pos: The starting point.
            parent: The parent for the starting position. If no font is given,
                this or one of its ancestors must implement :obj:`.HasMusicFont`.
            end_pos: The stopping point.
            end_parent: The parent for the ending position.
                If ``None``, defaults to ``self``.
            direction: The direction of the hairpin, where ``LEFT`` means diminuendo (>)
                and ``RIGHT`` means crescendo (<).
            width: The width of the wide hairpin. Defaults to 1 staff unit.
            font: If provided, this overrides any font found in the ancestor chain.
        """
        MusicPath.__init__(self, pos, parent, font=font, brush=Brush.no_brush())
        end_pos = Point.from_def(end_pos)
        Spanner2D.__init__(self, end_pos, end_parent or self)
        self.direction = direction
        self.width = width if width is not None else self.music_font.unit(1)
        self.pen = Pen(thickness=self.music_font.engraving_defaults["hairpinThickness"])
        self._draw_path()

    @property
    def music_font(self) -> MusicFont:
        return self._music_font

    @property
    def direction(self) -> DirectionX:
        """The direction of the hairpin.

        ``LEFT`` means diminuendo (>) and ``RIGHT`` means crescendo (<).
        """
        return self._direction

    @direction.setter
    def direction(self, value: DirectionX):
        self._direction = value

    def _find_hairpin_points(
        self,
    ) -> Tuple[
        Point, PositionedObject, Point, PositionedObject, Point, PositionedObject
    ]:
        """Find the hairpin path points for a set of parameters.

        The returned tuple is 3 pairs of Points and parents, where the
        outer 2 represent the wide ends of the hairpin and the middle
        represents the small end joint.
        """
        if self.direction == DirectionX.LEFT:
            joint_pos = self.end_pos
            joint_parent = self.end_parent
            end_center_pos = self.pos
            end_center_parent = self.parent
        else:
            joint_pos = self.pos
            joint_parent = self.parent
            end_center_pos = self.end_pos
            end_center_parent = self.end_parent
        dist = self.width / 2
        # Find relative distance from joint to end_center_pos
        parent_distance = joint_parent.map_to(end_center_parent)
        relative_stop = parent_distance + end_center_pos - joint_pos
        if relative_stop.y == ZERO:
            return (
                Point(end_center_pos.x, end_center_pos.y + dist),
                end_center_parent,
                joint_pos,
                joint_parent,
                Point(end_center_pos.x, end_center_pos.y - dist),
                end_center_parent,
            )
        elif relative_stop.x == ZERO:
            return (
                Point(end_center_pos.x + dist, end_center_pos.y),
                end_center_parent,
                joint_pos,
                joint_parent,
                Point(end_center_pos.x - dist, end_center_pos.y),
                end_center_parent,
            )
        # else ...

        # Find the two points (self.width / 2) away from the end_center_pos
        # which lie on the line perpendicular to the spanner line.

        #   Note that there is no risk of division by zero because
        #   previous if / elif statements catch those possibilities
        center_slope = relative_stop.y / relative_stop.x
        opening_slope = (center_slope * -1) ** -1
        opening_y_intercept = (end_center_pos.x * opening_slope) - end_center_pos.y
        # Find needed x coordinates of outer points
        #     x = dist / sqrt(1 + slope^2)
        first_x = end_center_pos.x + (dist / math.sqrt(1 + (opening_slope**2)))
        last_x = end_center_pos.x - (dist / math.sqrt(1 + (opening_slope**2)))
        # Calculate matching y coordinates from opening line function
        first_y = (first_x * opening_slope) - opening_y_intercept
        last_y = (last_x * opening_slope) - opening_y_intercept
        return (
            Point(first_x, first_y),
            end_center_parent,
            joint_pos,
            joint_parent,
            Point(last_x, last_y),
            end_center_parent,
        )

    def _draw_path(self):
        (
            first_pos,
            first_parent,
            mid_pos,
            mid_parent,
            last_pos,
            last_parent,
        ) = self._find_hairpin_points()
        self.move_to(first_pos.x, first_pos.y, first_parent)
        self.line_to(mid_pos.x, mid_pos.y, mid_parent)
        self.line_to(last_pos.x, last_pos.y, last_parent)



from __future__ import annotations

from typing import Optional

from neoscore.core.brush import BrushDef
from neoscore.core.directions import DirectionY
from neoscore.core.music_font import MusicFont
from neoscore.core.pen import PenDef
from neoscore.core.point import PointDef
from neoscore.core.positioned_object import PositionedObject
from neoscore.core.spanner import Spanner
from neoscore.core.units import Unit
from neoscore.western.abstract_slur import AbstractSlur


class Tie(Spanner, AbstractSlur):

    """A tie, conventionally drawn between notes.

    Note that this class currently provides no smart positioning logic to automatically
    attach to and span between noteheads in :obj:`.Chordrest`\ s.

    For slurs, see :obj:`.Slur`, which works identically to this class except that it
    is not constrainted to be horizontal.
    """

    def __init__(
        self,
        pos: PointDef,
        parent: PositionedObject,
        end_x: Unit,
        end_parent: Optional[PositionedObject] = None,
        direction: DirectionY = DirectionY.UP,
        height: Optional[Unit] = None,
        arch_length: Optional[Unit] = None,
        font: Optional[MusicFont] = None,
        brush: Optional[BrushDef] = None,
        pen: Optional[PenDef] = None,
    ):
        """
        Args:
            pos: The starting point.
            parent: The parent for the starting position. If no font is provided,
                this parent or one of its ancestors must implement :obj:`.HasStaffUnit`.
            end_x: The stopping point's x value. The y value will be automatically set
                so the tie is horizontal.
            direction: The vertical direction the slur arches.
            height: The ascent or descent of the curve given in absolute value.
                If omitted, a reasonable default is derived from other properties.
            arch_length: The x-offset of the outer ascent or descent curve control points.
                Smaller values give tighter arches. If omitted, a reasonable default is
                derived from other properties
            font: If provided, this overrides any font found in the ancestor chain.
            brush: The brush to fill shapes with.
            pen: The pen to draw outlines with.
        """
        AbstractSlur.__init__(
            self, pos, parent, direction, height, arch_length, font, brush, pen
        )
        Spanner.__init__(self, end_x, end_parent or self)
        self.draw_slur(self.end_pos, self.end_parent)



def running_in_ipython_gui_repl() -> bool:
    """Check if we are running inside an IPython REPL with the config gui=qt5.

    That config can be set either with ``ipython --gui=qt5 [-i script.py]``
    or by running the IPython magic command ``%gui qt5``.
    """
    try:
        import IPython  # type: ignore

        # None if not running within ipython
        ipython = IPython.get_ipython()
        if ipython:
            if ipython.config["TerminalIPythonApp"]["gui"] == "qt5":
                return True
    except:  # noqa
        pass
    return False



"""Custom hash functions for Qt objects.

These hash functions are mostly needed so Qt objects (or other classes including them)
can be hashed and used in cache dicts.
"""

from PyQt5.QtGui import QPainterPath, QTransform


def hash_transformed_path(path: QPainterPath, transform: QTransform) -> int:
    path_hash = hash_path(path)
    transform_hash = hash_transform(transform)
    return path_hash ^ transform_hash


# For some reason QTransform's qHash method seems to be omitted from PyQt5, so
# we implement our own based on the fields used in Qt's implementation here:
# github.com/qt/qtbase/blob/e05e3c776/src/gui/painting/qtransform.cpp#L778-L791
def hash_transform(t: QTransform) -> int:
    return hash(
        (
            t.m11(),
            t.m12(),
            t.m21(),
            t.m22(),
            t.dx(),
            t.dy(),
            t.m13(),
            t.m23(),
            t.m33(),
        )
    )


def hash_path(path: QPainterPath) -> int:
    h = 35891237 ^ path.fillRule()
    for i in range(path.elementCount()):
        el = path.elementAt(i)
        h ^= hash((el.type, el.x, el.y))
    return h
